#! /usr/bin/env python

from numpy.testing import assert_array_equal
import numpy as np

from landlab.testing import NumpyArrayTestingMixIn

import landlab.utils.structured_grid as sgrid
from landlab.grid.base import BAD_INDEX_VALUE, CLOSED_BOUNDARY


def test_node_x_2d():
    (x, _) = sgrid.node_coords((3, 2))

    assert_array_equal(x, np.array([0., 1.,
                                    0., 1.,
                                    0., 1., ]))


def test_node_x_2d_with_spacing():
    (x, _) = sgrid.node_coords((3, 2), (2., 10.))

    assert_array_equal(x, np.array([0., 10.,
                                    0., 10.,
                                    0., 10., ]))


def test_node_x_2d_with_origin():
    (x, _) = sgrid.node_coords((3, 2), (2., 10.), (-1., 1.))

    assert_array_equal(x, np.array([1., 11.,
                                    1., 11.,
                                    1., 11., ]))


def test_node_y_2d():
    (_, y) = sgrid.node_coords((3, 2))

    assert_array_equal(y, np.array([0., 0.,
                                    1., 1.,
                                    2., 2., ]))


def test_node_y_2d_with_spacing():
    (_, y) = sgrid.node_coords((3, 2), (2., 10.))

    assert_array_equal(y, np.array([0., 0.,
                                    2., 2.,
                                    4., 4., ]))


def test_node_y_2d_with_origin():
    (_, y) = sgrid.node_coords((3, 2), (2., 10.), (-1., 1.))

    assert_array_equal(y, np.array([-1., -1.,
                                    1.,  1.,
                                    3.,  3., ]))


def test_round_off_error():
    (x, y) = sgrid.node_coords((135, 127),
                               (5.4563957090392, 5.4563957090392),
                               (0., 0.))

    assert x.shape == (135 * 127, )
    assert y.shape == (135 * 127, )


def test_2d_shape_2_by_3():
    cell_nodes = sgrid.node_at_cell((2, 3))

    assert_array_equal(cell_nodes, np.array([]))


def test_2d_shape_3_by_3():
    cell_nodes = sgrid.node_at_cell((3, 3))

    assert_array_equal(cell_nodes, np.array([4]))


def test_shape_4_by_5():
    cell_nodes = sgrid.node_at_cell((4, 5))

    assert_array_equal(cell_nodes, np.array([6,  7,  8,
                                             11, 12, 13]))


# class TestGetNodeLinks(unittest.TestCase, NumpyArrayTestingMixIn):
def test_2d_3_by_2_from_links():
    (from_indices, _) = sgrid.node_index_at_link_ends((3, 2))

    assert_array_equal(from_indices, np.array([0, 1, 2, 3, 0, 2, 4]))


def test_2d_3_by_2_to_links():
    (_, to_indices) = sgrid.node_index_at_link_ends((3, 2))

    assert_array_equal(to_indices, np.array([2, 3, 4, 5, 1, 3, 5]))


def test_west_links():
    links = sgrid.west_links((3, 4))
    assert_array_equal(links, np.array([[-1,  8,  9, 10],
                                        [-1, 11, 12, 13],
                                        [-1, 14, 15, 16]]))

    links = sgrid.west_links((1, 4))
    assert_array_equal(links, np.array([[-1, 0, 1, 2]]))

    links = sgrid.west_links((4, 1))
    assert_array_equal(links, np.array([[-1], [-1], [-1], [-1]]))


def test_east_links():
    links = sgrid.east_links((3, 4))
    assert_array_equal(links, np.array([[8,  9, 10, -1],
                                        [11, 12, 13, -1],
                                        [14, 15, 16, -1]]))

    links = sgrid.east_links((1, 4))
    assert_array_equal(links, np.array([[0, 1, 2, -1]]))

    links = sgrid.east_links((4, 1))
    assert_array_equal(links, np.array([[-1, -1, -1, -1]]).T)


def test_north_links():
    links = sgrid.north_links((3, 4))
    assert_array_equal(links, np.array([[0,  1,  2,  3],
                                        [4,  5,  6,  7],
                                        [-1, -1, -1, -1]]))

    links = sgrid.north_links((1, 4))
    assert_array_equal(links, np.array([[-1, -1, -1, -1]]))

    links = sgrid.north_links((4, 1))
    assert_array_equal(links, np.array([[0, 1, 2, -1]]).T)


def test_south_links():
    links = sgrid.south_links((3, 4))
    assert_array_equal(links, np.array([[-1, -1, -1, -1],
                                        [0,  1,  2,  3],
                                        [4,  5,  6,  7]]))

    links = sgrid.south_links((1, 4))
    assert_array_equal(links, np.array([[-1, -1, -1, -1]]))

    links = sgrid.south_links((4, 1))
    assert_array_equal(links, np.array([[-1, 0, 1, 2]]).T)


def test_inlinks():
    links = sgrid.inlinks((3, 4))
    assert_array_equal(
        np.array([[-1, -1, -1, -1,  0,  1,  2,  3,  4,  5,  6,  7],
                  [-1,  8,  9, 10, -1, 11, 12, 13, -1, 14, 15, 16]]),
        links)


def test_outlinks():
    links = sgrid.outlinks((3, 4))
    assert_array_equal(
        np.array([[0,  1,  2,  3,  4,  5,  6,  7, -1, -1, -1, -1],
                  [8,  9, 10, -1, 11, 12, 13, -1, 14, 15, 16, -1]]),
        links)


# class TestNodeActiveCell(unittest.TestCase, NumpyArrayTestingMixIn):
def test_one_active_cell():
    active_cells = sgrid.active_cell_index_at_nodes((3, 3))

    assert_array_equal(
        active_cells,
        np.array([BAD_INDEX_VALUE, BAD_INDEX_VALUE, BAD_INDEX_VALUE,
                  BAD_INDEX_VALUE,               0, BAD_INDEX_VALUE,
                  BAD_INDEX_VALUE, BAD_INDEX_VALUE, BAD_INDEX_VALUE])
    )


def test_no_active_cells():
    active_cells = sgrid.active_cell_index_at_nodes((3, 2))

    assert_array_equal(
        active_cells,
        np.array([BAD_INDEX_VALUE, BAD_INDEX_VALUE,
                  BAD_INDEX_VALUE, BAD_INDEX_VALUE,
                  BAD_INDEX_VALUE, BAD_INDEX_VALUE])
    )


# class TestActiveCells(unittest.TestCase, NumpyArrayTestingMixIn):
def test_one_active_cell():
    active_cells = sgrid.active_cell_index((3, 3))

    assert_array_equal(active_cells, np.array([0]))


def test_no_active_cells():
    active_cells = sgrid.active_cell_index((3, 2))

    assert_array_equal(active_cells,
                       np.array([]))


# class TestCellCount(unittest.TestCase, NumpyArrayTestingMixIn):
def test_one_cell():
    n_cells = sgrid.cell_count((3, 3))
    assert n_cells == 1


def test_no_cells():
    n_cells = sgrid.cell_count((2, 3))
    assert n_cells == 0


# class TestInteriorCellCount(unittest.TestCase, NumpyArrayTestingMixIn):
def test_one_cell():
    n_cells = sgrid.interior_cell_count((3, 3))
    assert n_cells == 1


def test_no_cells():
    n_cells = sgrid.interior_cell_count((2, 3))
    assert n_cells == 0


# class TestActiveCellCount(unittest.TestCase, NumpyArrayTestingMixIn):
def test_one_cell():
    n_cells = sgrid.active_cell_count((3, 3))
    assert n_cells == 1


def test_no_cells():
    n_cells = sgrid.active_cell_count((2, 3))
    assert n_cells == 0


# class TestInteriorNodes(unittest.TestCase, NumpyArrayTestingMixIn):
def test_4_by_5():
    interiors = sgrid.interior_nodes((4, 5))
    assert_array_equal(interiors, np.array([6, 7, 8, 11, 12, 13]))


def test_no_interiors():
    interiors = sgrid.interior_nodes((2, 3))
    assert_array_equal(interiors, np.array([]))


# class TestNodeStatus(unittest.TestCase, NumpyArrayTestingMixIn):
def test_4_by_5():
    status = sgrid.stat_at_node((4, 5))
    assert status.dtype == np.int8
    assert_array_equal(status,
                       np.array([1, 1, 1, 1, 1,
                                 1, 0, 0, 0, 1,
                                 1, 0, 0, 0, 1,
                                 1, 1, 1, 1, 1, ]))


def test_no_interiors():
    status = sgrid.status_at_node((2, 3))
    assert status.dtype == np.int8
    assert_array_equal(status,
                       np.array([1, 1, 1,
                                 1, 1, 1, ]))


# class TestActiveLinks(unittest.TestCase, NumpyArrayTestingMixIn):
"""
*--27-->*--28-->*--29-->*--30-->*
^       ^       ^       ^       ^
10      11      12      13      14
|       |       |       |       |
*--23-->*--24-->*--25-->*--26-->*
^       ^       ^       ^       ^
5       6       7       8       9
|       |       |       |       |
*--19-->*--20-->*--21-->*--22-->*
^       ^       ^       ^       ^
0       1       2       3       4
|       |       |       |       |
*--15-->*--16-->*--17-->*--18-->*
"""


def test_4_by_5():
    active_links = sgrid.active_links((4, 5))
    assert_array_equal(active_links,
                       np.array([1, 2, 3, 6, 7, 8, 11, 12, 13,
                                 19, 20, 21, 22, 23, 24, 25, 26]))
    assert len(active_links) == sgrid.active_link_count((4, 5))


def test_with_status_at_node():
    status = sgrid.status_at_node((4, 5))
    status[6] = sgrid.CLOSED_BOUNDARY
    active_links = sgrid.active_links((4, 5), node_status_array=status)

    assert_array_equal(active_links,
                       np.array([2, 3, 7, 8, 11, 12, 13,
                                 21, 22, 23, 24, 25, 26]))


def test_with_link_nodes():
    link_nodes = sgrid.node_index_at_link_ends((4, 5))
    active_links = sgrid.active_links((4, 5), link_nodes=link_nodes)

    assert_array_equal(active_links,
                       np.array([1, 2, 3, 6, 7, 8, 11, 12, 13,
                                 19, 20, 21, 22, 23, 24, 25, 26]))
    assert len(active_links) == sgrid.active_link_count((4, 5))


def test_vertical_active_link_count():
    link_count = sgrid.vertical_active_link_count((3, 4))
    assert 4 == link_count

    link_count = sgrid.vertical_active_link_count((3, 2))
    assert 0 == link_count

    node_status = np.ones((4, 5), dtype=np.int)
    link_count = sgrid.vertical_active_link_count((4, 5),
                                                  node_status=node_status)
    assert 9 == link_count

    link_count = sgrid.vertical_active_link_count((4, 5),
                                                  node_status=node_status)
    node_status[0, 1] = 0
    link_count = sgrid.vertical_active_link_count((4, 5),
                                                  node_status=node_status)
    assert 8 == link_count

    node_status[2, 1] = 0
    link_count = sgrid.vertical_active_link_count((4, 5),
                                                  node_status=node_status)
    assert 6 == link_count

    node_status[2, 2] = 0
    link_count = sgrid.vertical_active_link_count((4, 5),
                                                  node_status=node_status)
    assert 4 == link_count

    node_status[1, 1] = 0
    link_count = sgrid.vertical_active_link_count((4, 5),
                                                  node_status=node_status)
    assert 4 == link_count


def test_horizontal_active_link_count():
    link_count = sgrid.horizontal_active_link_count((3, 4))
    assert 3 == link_count

    link_count = sgrid.horizontal_active_link_count((2, 3))
    assert 0 == link_count

    node_status = np.ones((4, 5), dtype=np.int)
    link_count = sgrid.horizontal_active_link_count(
        (4, 5), node_status=node_status)
    assert 8 == link_count

    link_count = sgrid.horizontal_active_link_count(
        (4, 5), node_status=node_status)
    node_status[0, 1] = 0
    link_count = sgrid.horizontal_active_link_count((4, 5),
                                                    node_status=node_status)
    assert 8 == link_count

    node_status[2, 1] = 0
    link_count = sgrid.horizontal_active_link_count((4, 5),
                                                    node_status=node_status)
    assert 6 == link_count

    node_status[2, 2] = 0
    link_count = sgrid.horizontal_active_link_count((4, 5),
                                                    node_status=node_status)
    assert 5 == link_count

    node_status[1, 1] = 0
    link_count = sgrid.horizontal_active_link_count((4, 5),
                                                    node_status=node_status)
    assert 3 == link_count


def test_horizontal_active_link_ids():
    links = sgrid.horizontal_active_link_ids((3, 4))
    assert_array_equal(links, np.array([[4, 5, 6]]))

    links = sgrid.horizontal_active_link_ids((1, 4))
    expected = np.array([], ndmin=2, dtype=np.int64)
    expected.shape = (0, 3)
    assert_array_equal(expected, links)

    links = sgrid.horizontal_active_link_ids((4, 1))
    expected.shape = (2, 0)
    assert_array_equal(expected, links)

    node_status = np.ones((4, 5), dtype=int)
    links = sgrid.horizontal_active_link_ids((4, 5),
                                             node_status=node_status)
    assert_array_equal(links, np.array([[9, 10, 11, 12],
                                        [13, 14, 15, 16]]))

    node_status = np.ones((4, 5), dtype=int)
    node_status[1, 1] = 0
    links = sgrid.horizontal_active_link_ids((4, 5),
                                             node_status=node_status)
    assert_array_equal(links, np.array([[-1, -1,  7,  8],
                                        [9, 10, 11, 12]]))

    node_status[2, 1] = 0
    links = sgrid.horizontal_active_link_ids((4, 5),
                                             node_status=node_status)
    assert_array_equal(links, np.array([[-1, -1, 6, 7],
                                        [-1, -1, 8, 9]]))

    node_status[0, 0] = 0
    links = sgrid.horizontal_active_link_ids((4, 5),
                                             node_status=node_status)
    assert_array_equal(links, np.array([[-1, -1, 6, 7],
                                        [-1, -1, 8, 9]]))


def test_vertical_active_link_ids():
    links = sgrid.vertical_active_link_ids((3, 4))
    assert_array_equal(links, np.array([[0, 1], [2, 3]]))

    links = sgrid.vertical_active_link_ids((1, 4))
    expected = np.array([], ndmin=2, dtype=np.int64)
    expected.shape = (0, 2)
    assert_array_equal(expected, links)

    links = sgrid.vertical_active_link_ids((4, 1))
    expected.shape = (3, 0)
    assert_array_equal(expected, links)

    node_status = np.ones((4, 5), dtype=int)
    links = sgrid.vertical_active_link_ids((4, 5), node_status=node_status)
    assert_array_equal(links, np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]]))

    node_status = np.ones((4, 5), dtype=int)
    node_status[1, 1] = 0
    links = sgrid.vertical_active_link_ids((4, 5), node_status=node_status)
    assert_array_equal(links,
                       np.array([[-1, 0, 1], [-1, 2, 3], [4, 5, 6]]))

    node_status[2, 1] = 0
    links = sgrid.vertical_active_link_ids((4, 5), node_status=node_status)
    assert_array_equal(links,
                       np.array([[-1, 0, 1], [-1, 2, 3], [-1, 4, 5]]))

    node_status[0, 0] = 0
    links = sgrid.vertical_active_link_ids((4, 5), node_status=node_status)
    assert_array_equal(links,
                       np.array([[-1, 0, 1], [-1, 2, 3], [-1, 4, 5]]))


def test_west_links():
    links = sgrid.active_west_links((3, 4))
    assert_array_equal(links, np.array([[-1, -1, -1, -1],
                                        [-1,  4,  5,  6],
                                        [-1, -1, -1, -1]]))

    links = sgrid.active_west_links((1, 4))
    assert_array_equal(links, np.array([[-1, -1, -1, -1]]))

    links = sgrid.active_west_links((4, 1))
    assert_array_equal(links, np.array([[-1, -1, -1, -1]]).T)


def test_east_links():
    links = sgrid.active_east_links((3, 4))
    assert_array_equal(links, np.array([[-1, -1, -1, -1],
                                        [4,  5,  6, -1],
                                        [-1, -1, -1, -1]]))

    links = sgrid.active_east_links((1, 4))
    assert_array_equal(links, np.array([[-1, -1, -1, -1]]))

    links = sgrid.active_east_links((4, 1))
    assert_array_equal(links, np.array([[-1, -1, -1, -1]]).T)

    links = sgrid.horizontal_active_link_ids((4, 5))
    assert_array_equal(np.array([[9, 10, 11, 12],
                                 [13, 14, 15, 16]]),
                       links)

    links = sgrid.active_east_links((4, 5))
    assert_array_equal(np.array([[-1, -1, -1, -1, -1],
                                 [9, 10, 11, 12, -1],
                                 [13, 14, 15, 16, -1],
                                 [-1, -1, -1, -1, -1]]),
                       links)


def test_north_links():
    links = sgrid.active_north_links((3, 4))
    assert_array_equal(links, np.array([[-1,  0,  1, -1],
                                        [-1,  2,  3, -1],
                                        [-1, -1, -1, -1]]))

    links = sgrid.active_north_links((1, 4))
    assert_array_equal(links, np.array([[-1, -1, -1, -1]]))

    links = sgrid.active_north_links((4, 1))
    assert_array_equal(links, np.array([[-1, -1, -1, -1]]).T)


def test_south_links():
    links = sgrid.active_south_links((3, 4))
    assert_array_equal(links, np.array([[-1, -1, -1, -1],
                                        [-1,  0,  1, -1],
                                        [-1,  2,  3, -1]]))

    links = sgrid.active_south_links((1, 4))
    assert_array_equal(links, np.array([[-1, -1, -1, -1]]))

    links = sgrid.active_south_links((4, 1))
    assert_array_equal(links, np.array([[-1, -1, -1, -1]]).T)


def test_inlinks():
    links = sgrid.active_inlinks((3, 4))
    assert_array_equal(
        np.array([[-1, -1, -1, -1, -1,  0,  1, -1, -1,  2,  3, -1],
                  [-1, -1, -1, -1, -1,  4,  5,  6, -1, -1, -1, -1]]),
        links)


def test_outlinks():
    links = sgrid.active_outlinks((3, 4))
    assert_array_equal(
        np.array([[-1,  0,  1, -1, -1,  2,  3, -1, -1, -1, -1, -1],
                  [-1, -1, -1, -1,  4,  5,  6, -1, -1, -1, -1, -1]]),
        links)


def test_outlinks_4x5():
    links = sgrid.active_outlinks((4, 5))

    assert_array_equal(np.array([[-1,  0,  1,  2, -1,
                                  -1,  3,  4,  5, -1,
                                  -1,  6,  7,  8, -1,
                                  -1, -1, -1, -1, -1],
                                 [-1, -1, -1, -1, -1,
                                  9, 10, 11, 12, -1,
                                  13, 14, 15, 16, -1,
                                  -1, -1, -1, -1, -1]]),
                       links)


def test_inlinks_4x5():
    links = sgrid.active_inlinks((4, 5))

    assert_array_equal(np.array([[-1, -1, -1, -1, -1,
                                  -1,  0,  1,  2, -1,
                                  -1,  3,  4,  5, -1,
                                  -1,  6, 7,  8, -1],
                                 [-1, -1, -1, -1, -1,
                                  -1,  9, 10, 11, 12,
                                  -1, 13, 14, 15, 16,
                                  -1, -1, -1, -1, -1]]),
                       links)


# class TestFaces(unittest.TestCase, NumpyArrayTestingMixIn):
def test_face_count():
    assert 17 == sgrid.face_count((4, 5))
    assert 4 == sgrid.face_count((3, 3))
    assert 0 == sgrid.face_count((2, 100))
    assert 0 == sgrid.face_count((100, 2))
    assert 0 == sgrid.face_count((100, 1))


def test_active_face_count():
    assert 17 == sgrid.active_face_count((4, 5))
    assert 4 == sgrid.active_face_count((3, 3))
    assert 0 == sgrid.active_face_count((2, 100))
    assert 0 == sgrid.active_face_count((100, 2))
    assert 0 == sgrid.active_face_count((100, 1))


def test_active_faces():
    active_faces = sgrid.active_face_index((4, 5))
    assert_array_equal(np.array([0,  1,  2,
                                 3,  4,  5,
                                 6,  7,  8,
                                 9, 10, 11, 12,
                                 13, 14, 15, 16]),
                       active_faces)


# class TestLinkFaces(unittest.TestCase, NumpyArrayTestingMixIn):
def test_4_by_5():
    link_faces = sgrid.face_at_link((4, 5))

    BAD = sgrid.BAD_INDEX_VALUE

    assert_array_equal(link_faces, np.array([BAD, 0, 1, 2, BAD,
                                             BAD, 3, 4, 5, BAD,
                                             BAD, 6, 7, 8, BAD,
                                             BAD, BAD, BAD, BAD,
                                             9, 10, 11, 12,
                                             13, 14, 15, 16,
                                             BAD, BAD, BAD, BAD]))


def test_with_active_links():
    active_links = sgrid.active_links((4, 5))
    active_links = active_links[:-1]
    link_faces = sgrid.face_at_link((4, 5), actives=active_links)

    BAD = sgrid.BAD_INDEX_VALUE

    assert_array_equal(link_faces, np.array([BAD, 0, 1, 2, BAD,
                                             BAD, 3, 4, 5, BAD,
                                             BAD, 6, 7, 8, BAD,
                                             BAD, BAD, BAD, BAD,
                                             9, 10, 11, 12,
                                             13, 14, 15, BAD,
                                             BAD, BAD, BAD, BAD]))


# class TestReshapeArray(unittest.TestCase, NumpyArrayTestingMixIn):
def test_default():
    x = np.arange(12.)
    y = sgrid.reshape_array((3, 4), x)

    assert y.shape == (3, 4)
    assert_array_equal(x, y.flat)
    assert y.flags['C_CONTIGUOUS']
    assert y.base is x


def test_copy():
    x = np.arange(12.)
    y = sgrid.reshape_array((3, 4), x, copy=True)

    assert y.shape == (3, 4)
    assert_array_equal(x, y.flat)
    assert y.flags['C_CONTIGUOUS']
    assert y.base is None

    y[0][0] = 0.
    assert_array_equal(x, np.array([0.,  1.,  2., 3.,
                                    4.,  5.,  6., 7.,
                                    8.,  9., 10., 11.]))


def test_flip():
    x = np.arange(12.)
    y = sgrid.reshape_array((3, 4), x, flip_vertically=True)

    assert y.shape == (3, 4)
    assert_array_equal(y, np.array([[8.,  9., 10., 11.],
                                    [4.,  5.,  6., 7.],
                                    [0.,  1.,  2., 3.]]))
    assert not y.flags['C_CONTIGUOUS']
    assert y.base is not None

    y[0][0] = 0.
    assert_array_equal(x, np.array([0.,  1.,  2., 3.,
                                    4.,  5.,  6., 7.,
                                    0.,  9., 10., 11.]))


def test_flip_copy():
    x = np.arange(12.)
    y = sgrid.reshape_array((3, 4), x, flip_vertically=True, copy=True)

    assert y.shape == (3, 4)
    assert_array_equal(y, np.array([[8.,  9., 10., 11.],
                                    [4.,  5.,  6., 7.],
                                    [0.,  1.,  2., 3.]]))
    assert y.flags['C_CONTIGUOUS']
    assert y.base is not x


# class TestDiagonalArray(unittest.TestCase, NumpyArrayTestingMixIn):
def test_default():
    diags = sgrid.diagonal_node_array((2, 3), out_of_bounds=-1)
    assert_array_equal(diags,
                       np.array([[4, -1, -1, -1],
                                 [5,  3, -1, -1],
                                 [-1,  4, -1, -1],
                                 [-1, -1, -1,  1],
                                 [-1, -1,  0,  2],
                                 [-1, -1,  1, -1]]))

    assert diags.base is None
    assert diags.flags['C_CONTIGUOUS']


def test_non_contiguous():
    diags = sgrid.diagonal_node_array((2, 3), out_of_bounds=-1,
                                      contiguous=False)
    assert_array_equal(diags,
                       np.array([[4, -1, -1, -1],
                                 [5,  3, -1, -1],
                                 [-1,  4, -1, -1],
                                 [-1, -1, -1,  1],
                                 [-1, -1,  0,  2],
                                 [-1, -1,  1, -1]]))

    assert isinstance(diags.base, np.ndarray)
    assert not diags.flags['C_CONTIGUOUS']


def test_boundary_node_mask_no_actives():
    diags = sgrid.diagonal_node_array((2, 3), out_of_bounds=-1,
                                      boundary_node_mask=-2)
    assert_array_equal(diags, - 2 * np.ones((6, 4)))


def test_boundary_node_mask():
    diags = sgrid.diagonal_node_array((3, 3), out_of_bounds=-1,
                                      boundary_node_mask=-2)
    assert_array_equal(diags,
                       np.array([[-2, -2, -2, -2], [-2, -2, -2, -2], [-2, -2, -2, -2],
                                 [-2, -2, -2, -2], [8,  6,
                                                    0,  2], [-2, -2, -2, -2],
                                 [-2, -2, -2, -2], [-2, -2, -2, -2], [-2, -2, -2, -2]]))


# class TestNeighborArray(unittest.TestCase, NumpyArrayTestingMixIn):
def test_default():
    neighbors = sgrid.neighbor_node_array((2, 3))

    BAD = sgrid.BAD_INDEX_VALUE

    assert_array_equal(
        neighbors,
        np.array([[1,   3, BAD, BAD],
                  [2,   4,   0, BAD],
                  [BAD,   5,   1, BAD],
                  [4, BAD, BAD,   0],
                  [5, BAD,   3,   1],
                  [BAD, BAD,   4,   2]]).T)

    assert neighbors.flags['C_CONTIGUOUS']
    assert neighbors.base is None


def test_set_out_of_bounds():
    neighbors = sgrid.neighbor_node_array((2, 3), inactive=-1)
    assert_array_equal(neighbors,
                       np.array([[1,  3, -1, -1],
                                 [2,  4,  0, -1],
                                 [-1,  5,  1, -1],
                                 [4, -1, -1,  0],
                                 [5, -1,  3,  1],
                                 [-1, -1,  4,  2]]).T)


# class TestInlinkMatrix(unittest.TestCase, NumpyArrayTestingMixIn):
def test_no_inactive():
    inlinks = sgrid.setup_active_inlink_matrix((4, 5), return_count=False)
    assert_array_equal(inlinks,
                       np.array([[-1, -1, -1, -1, -1,
                                  -1,  0,  1,  2, -1,
                                  -1,  3,  4,  5, -1,
                                  -1,  6,  7,  8, -1],
                                 [-1, -1, -1, -1, -1,
                                  -1,  9, 10, 11, 12,
                                  -1, 13, 14, 15, 16,
                                  -1, -1, -1, -1, -1]]))


def test_inactive():
    status = np.ones((4, 5))
    status[1, 1] = 0
    inlinks = sgrid.setup_active_inlink_matrix((4, 5), return_count=False,
                                               node_status=status)
    assert_array_equal(inlinks,
                       np.array([[-1, -1, -1, -1, -1,
                                  -1, -1,  0,  1, -1,
                                  -1, -1,  2,  3, -1,
                                  -1,  4,  5,  6, -1],
                                 [-1, -1, -1, -1, -1,
                                  -1, -1, -1,  7,  8,
                                  -1,  9, 10, 11, 12,
                                  -1, -1, -1, -1, -1]]))


def test_out_link_ids_at_nodes():
    links_ids = sgrid.outlink_index_at_node((4, 5))
    assert_array_equal(
        np.array([[ 4,  5,  6,  7,  8,
                   13, 14, 15, 16, 17,
                   22, 23, 24, 25, 26,
                   -1, -1, -1, -1, -1],
                  [ 0,  1,  2,  3, -1,
                    9, 10, 11, 12, -1,
                   18, 19, 20, 21, -1,
                   27, 28, 29, 30, -1]]),
        links_ids)


def test_in_link_ids_at_nodes():
    links_ids = sgrid.inlink_index_at_node((4, 5))
    assert_array_equal(
        np.array([[-1, -1, -1, -1, -1,  
                    4,  5,  6,  7,  8,
                   13, 14, 15, 16, 17, 
                   22, 23, 24, 25, 26],
                  [-1,  0,  1,  2,  3, 
                   -1,  9, 10, 11, 12,
                   -1, 18, 19, 20, 21,
                   -1, 27, 28, 29, 30]]),
        links_ids)



"""Store arrays of variable-length arrays implemented with masked arrays.

Implements a JaggedArray class using numpy masked arrays.

Examples
--------

Create a JaggedArray that stores link IDs for the links attached to the
nodes of a 3x3 grid.

>>> from landlab.utils.jaggedarray_ma import JaggedArray
>>> links_at_node = JaggedArray([
...     [0, 6],
...     [1, 7, 0],
...     [8, 1],
...     [2, 9, 6],
...     [3, 10, 2, 7],
...     [11, 3, 8],
...     [4, 7],
...     [5, 10, 4],
...     [5, 11]])

Make up some data that provides values at each of the links.

>>> value_at_link = np.arange(12, dtype=float)

Create another JaggedArray. Here we store the values at each of the links
attached to nodes of the grid.

>>> values_at_node = JaggedArray.empty_like(links_at_node, dtype=float)
>>> values_at_node.array = value_at_link[links_at_node.array]

Now operate on the link values for each node.

>>> values_at_node.foreach_row(np.sum)
array([  6.,   8.,   9.,  17.,  22.,  22.,  11.,  19.,  16.])
>>> values_at_node.foreach_row(np.min)
array([ 0.,  0.,  1.,  2.,  2.,  3.,  4.,  4.,  5.])
>>> values_at_node.foreach_row(np.ptp)
array([ 6.,  7.,  7.,  7.,  8.,  8.,  3.,  6.,  6.])
"""
import numpy as np
from six.moves import range


class JaggedArray(object):

    """
    A container for an array of variable-length arrays.

    JaggedArray([row0, row1, ...])
    JaggedArray(values, values_per_row)

    Examples
    --------
    Create a JaggedArray with an array of arrays.

    >>> from landlab.utils.jaggedarray_ma import JaggedArray
    >>> x = JaggedArray([[0, 1, 2], [3, 4]])
    >>> x.array
    array([0, 1, 2, 3, 4])

    Create a JaggedArray as a 1D array and a list or row lengths.

    >>> x = JaggedArray([0, 1, 2, 3, 4], (3, 2))
    >>> x.array
    array([0, 1, 2, 3, 4])
    """

    def __init__(self, *args):
        """
        JaggedArray([row0, row1, ...])
        JaggedArray(values, values_per_row)

        Examples
        --------
        Create a JaggedArray with an array of arrays.

        >>> from landlab.utils.jaggedarray_ma import JaggedArray
        >>> x = JaggedArray([[0, 1, 2], [3, 4]])
        >>> x.array
        array([0, 1, 2, 3, 4])

        Create a JaggedArray as a 1D array and a list or row lengths.

        >>> x = JaggedArray([0, 1, 2, 3, 4], (3, 2))
        >>> x.array
        array([0, 1, 2, 3, 4])
        """
        if len(args) == 1:
            if isinstance(args[0], np.ma.core.MaskedArray):
                mat = args[0]
            else:
                mat = JaggedArray.ma_from_list_of_lists(args[0])
        else:
            mat = JaggedArray.ma_from_flat_array(args[0], args[1])

        self._values = mat
        self._number_of_rows = mat.shape[0]

    @staticmethod
    def ma_from_list_of_lists(rows, dtype=None):
        """Create a masked array from a list of lists.

        Parameters
        ----------
        rows : array_like or array_like
            Rows of the jagged array.
        dtype : np.dtype, optional
            The data type of the new masked array.

        Returns
        -------
        np.masked_array
            A new masked array.
        """
        values_per_row = [len(row) for row in rows]
        mat = np.ma.masked_all((len(rows), max(values_per_row)),
                               dtype=dtype or int)
        for (row_number, row) in enumerate(rows):
            mat[row_number, :len(row)] = row

        return mat

    @staticmethod
    def ma_from_flat_array(array, values_per_row):
        """Create a masked array from a flat array.

        Parameters
        ----------
        array : array_like
            Values of the jagged array.
        values_per_row : array_like of int
            Number of values in each row of the jagged array.

        Returns
        -------
        np.masked_array
            A new masked array.
        """
        array = np.array(array)
        mat = np.ma.masked_all((len(values_per_row), max(values_per_row)),
                               dtype=array.dtype)
        offset = 0
        for row_number in range(mat.shape[0]):
            n_valid = values_per_row[row_number]
            mat[row_number, :n_valid] = array[offset:offset + n_valid]
            offset += n_valid

        return mat

    @property
    def array(self):
        """The jagged array as a 1D array.

        Returns
        -------
        array :
            A view of the underlying 1D array.

        Examples
        --------
        >>> from landlab.utils.jaggedarray_ma import JaggedArray
        >>> x = JaggedArray([[0, 1, 2], [3, 4]])
        >>> x.array
        array([0, 1, 2, 3, 4])

        >>> x.array = np.array([1, 1, 2, 3, 4])
        >>> x.array
        array([1, 1, 2, 3, 4])
        """
        return self._values.compressed()

    @property
    def masked_array(self):
        """The jagged array as a masked array.

        Returns
        -------
        np.masked_array :
            The underlying masked array.
        """
        return self._values

    @array.setter
    def array(self, array):
        """Set the data of the jagged array from a 1D array.

        Parameters
        ----------
        array : array_like
            The new values of the array.
        """
        self._values[~ self._values.mask] = array

    @property
    def size(self):
        """Number of array elements.

        Returns
        -------
        int :
            Number of values in the array.

        Examples
        --------
        >>> from landlab.utils.jaggedarray_ma import JaggedArray
        >>> x = JaggedArray([[0, 1, 2], [3, 4]])
        >>> x.size
        5
        """
        return self.array.size

    @property
    def number_of_rows(self):
        """Number of array rows.

        Returns
        -------
        int :
            Number of rows in the array.

        Examples
        --------
        >>> from landlab.utils.jaggedarray_ma import JaggedArray
        >>> x = JaggedArray([[0, 1, 2], [3, 4]])
        >>> x.number_of_rows == 2
        True
        """
        return self._number_of_rows

    @staticmethod
    def _offsets_from_values_per_row(values_per_row):
        """Get offsets into the base array from array lengths.

        Parameters
        ----------
        values_per_row : array of int
            The number of values in each row of the JaggedArray.

        Returns
        -------
        ndarray
            An array of offsets.
        """
        offset = np.empty(len(values_per_row) + 1, dtype=int)
        np.cumsum(values_per_row, out=offset[1:])
        offset[0] = 0
        return offset

    @staticmethod
    def empty_like(jagged, dtype=None):
        """Create a new JaggedArray that is like another one.

        Parameters
        ----------
        jagged : JaggedArray
            A JaggedArray to copy.
        dtype : np.dtype
            The data type of the new JaggedArray.

        Returns
        -------
        JaggedArray
            A new JaggedArray.
        """
        return JaggedArray(np.ma.empty_like(jagged.masked_array, dtype=dtype))

    def length_of_row(self, row):
        """Number of values in a given row.

        Parameters
        ----------
        row : int
            Index to a row.

        Returns
        -------
        int :
            Number of values in the row.

        Examples
        --------
        >>> from landlab.utils.jaggedarray_ma import JaggedArray
        >>> x = JaggedArray([[0, 1, 2], [3, 4]])
        >>> x.length_of_row(0)
        3
        >>> x.length_of_row(1)
        2
        """
        return len(self.row(row))

    def row(self, row):
        """Get the values of a row as an array.

        Parameters
        ----------
        row : int
            Index to a row.

        Returns
        -------
        array :
            Values in the row as a slice of the underlying array.

        Examples
        --------
        >>> from landlab.utils.jaggedarray_ma import JaggedArray
        >>> x = JaggedArray([[0, 1, 2], [3, 4]])
        >>> x.row(0)
        array([0, 1, 2])
        >>> x.row(1)
        array([3, 4])
        """
        return self._values[row].compressed()

    def __iter__(self):
        """Iterate over the rows of the array.

        Examples
        --------
        >>> from landlab.utils.jaggedarray_ma import JaggedArray
        >>> x = JaggedArray([[0, 1, 2], [3, 4]])
        >>> for row in x: row
        array([0, 1, 2])
        array([3, 4])
        """
        for row in self._values:
            yield row.compressed()

    def foreach_row(self, func, out=None):
        """Apply an operator row-by-row.

        Examples
        --------
        >>> from landlab.utils.jaggedarray_ma import JaggedArray
        >>> x = JaggedArray([[0, 1, 2], [3, 4]])
        >>> x.foreach_row(np.sum)
        array([3, 7])

        >>> out = np.empty(2, dtype=int)
        >>> x.foreach_row(np.sum, out=out) is out
        True
        >>> out
        array([3, 7])
        """
        if out is None:
            return func(self._values, axis=1).compressed()
        else:
            return func(self._values, axis=1, out=out)



import pytest
import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid
from landlab.grid.base import BAD_INDEX_VALUE as X


def test_all_active_neighbors():
    rmg = RasterModelGrid(5, 4)
    expected = np.array([
        [X, X, X, X], [X, 5, X, X], [X, 6, X, X], [X, X, X, X],
        [5, X, X, X], [6, 9, 4, 1], [7, 10, 5, 2], [X, X, 6, X],
        [9, X, X, X], [10, 13, 8, 5], [11, 14, 9, 6], [X, X, 10, X],
        [13, X, X, X], [14, 17, 12, 9], [15, 18, 13, 10], [X, X, 14, X],
        [X, X, X, X], [X, X, X, 13], [X, X, X, 14], [X, X, X, X],
    ])
    assert_array_equal(rmg.active_adjacent_nodes_at_node, expected)


def test_all_neighbors():
    rmg = RasterModelGrid((5, 4))
    expected = np.array([
        [ 1,  4, X,  X], [ 2,  5,  0,  X], [ 3,  6,  1,  X], [X,  7,  2,  X],
        [ 5,  8, X,  0], [ 6,  9,  4,  1], [ 7, 10,  5,  2], [X, 11,  6,  3],
        [ 9, 12, X,  4], [10, 13,  8,  5], [11, 14,  9,  6], [X, 15, 10,  7],
        [13, 16, X,  8], [14, 17, 12,  9], [15, 18, 13, 10], [X, 19, 14, 11],
        [17,  X, X, 12], [18,  X, 16, 13], [19,  X, 17, 14], [X,  X, 18, 15],
    ])
    assert_array_equal(rmg.adjacent_nodes_at_node, expected)


def test_active_neighbor_list_with_scalar_arg():
    rmg = RasterModelGrid(5, 4)

    assert_array_equal(rmg.active_adjacent_nodes_at_node[6],
                       np.array([7, 10, 5, 2]))
    assert_array_equal(rmg.active_adjacent_nodes_at_node[-1],
                       np.array([X, X, X, X]))
    assert_array_equal(rmg.active_adjacent_nodes_at_node[-2],
                       np.array([X, X, X, 14]))


def test_neighbor_list_with_scalar_arg():
    rmg = RasterModelGrid((5, 4))

    assert_array_equal(rmg.adjacent_nodes_at_node[6], np.array([7, 10, 5, 2]))
    assert_array_equal(rmg.adjacent_nodes_at_node[-1], np.array([X, X, 18, 15]))
    assert_array_equal(rmg.adjacent_nodes_at_node[-2], np.array([19, X, 17, 14]))


def test_active_neighbor_list_with_array_arg():
    rmg = RasterModelGrid(5, 4)
    assert_array_equal(rmg.active_adjacent_nodes_at_node[[6, -1]],
                       np.array([[7, 10, 5, 2], [X, X, X, X]]))


def test_neighbor_list_with_array_arg():
    rmg = RasterModelGrid(5, 4)
    assert_array_equal(rmg.adjacent_nodes_at_node[(6, -1), :],
                       np.array([[7, 10, 5, 2], [X, X, 18, 15]]))


def test_neighbor_list_is_read_only():
    rmg = RasterModelGrid(5, 4)
    with pytest.raises(ValueError):
        rmg.adjacent_nodes_at_node[0] = [1, 2, 3, 4]


def test_neighbors_is_contiguous():
    rmg = RasterModelGrid(5, 4)
    assert rmg.adjacent_nodes_at_node.flags['C_CONTIGUOUS']


def test_active_neighbor_list_boundary():
    """All of the neighbor IDs for a boundary cell are -1."""
    rmg = RasterModelGrid(5, 4)
    import landlab.utils.structured_grid as sgrid
    rmg.set_closed_nodes([0, 1, 2, 3, 4, 7, 8, 11, 12, 15, 16, 17, 18, 19])

    for node_id in sgrid.boundary_iter(rmg.shape):
        assert_array_equal(rmg.active_adjacent_nodes_at_node[node_id],
                           np.array([X, X, X, X]))


def test_all_diagonals():
    rmg = RasterModelGrid((5, 4))
    expected = np.array([
        [ 5,  X, X,  X], [ 6,  4,  X,  X], [ 7,  5,  X,  X], [X,  6,  X,  X],
        [ 9,  X, X,  1], [10,  8,  0,  2], [11,  9,  1,  3], [X, 10,  2,  X],
        [13,  X, X,  5], [14, 12,  4,  6], [15, 13,  5,  7], [X, 14,  6,  X],
        [17,  X, X,  9], [18, 16,  8, 10], [19, 17,  9, 11], [X, 18, 10,  X],
        [ X,  X, X, 13], [ X,  X, 12, 14], [ X,  X, 13, 15], [X,  X, 14,  X],
    ])
    assert_array_equal(rmg.diagonal_adjacent_nodes_at_node, expected)


def test_diagonal_list_with_scalar_arg():
    rmg = RasterModelGrid((5, 4))

    assert_array_equal(rmg.diagonal_adjacent_nodes_at_node[6], np.array([11, 9, 1, 3]))
    assert_array_equal(rmg.diagonal_adjacent_nodes_at_node[-1], np.array([X, X, 14, X]))
    assert_array_equal(rmg.diagonal_adjacent_nodes_at_node[-2], np.array([X, X, 13, 15]))


def test_diagonal_list_with_array_arg():
    rmg = RasterModelGrid(5, 4)
    assert_array_equal(rmg.diagonal_adjacent_nodes_at_node[(6, -1), :],
                       np.array([[11, 9, 1, 3], [X, X, 14, X]]))


def test_diagonal_list_is_read_only():
    rmg = RasterModelGrid(5, 4)
    with pytest.raises(ValueError):
        rmg.diagonal_adjacent_nodes_at_node[0] = [1, 2, 3, 4]


def test_diagonals_is_contiguous():
    rmg = RasterModelGrid(5, 4)
    assert rmg.diagonal_adjacent_nodes_at_node.flags['C_CONTIGUOUS']



import numpy as np
cimport numpy as np
cimport cython


from libc.stdlib cimport malloc, free
# from libc.stdlib cimport mergesort, qsort


ctypedef np.int_t INT_t
ctypedef np.float_t FLOAT_t

cdef struct Sorter:
    INT_t index
    FLOAT_t value


cdef struct IntSorter:
    INT_t index
    INT_t value


cdef extern from "stdlib.h":
    ctypedef void const_void "const void"
    int qsort(void *, size_t, size_t,
               int(*)(const_void *, const_void *)) nogil
    int	mergesort(void *, size_t, size_t,
                  int (*)(const_void *, const_void *)) nogil


cdef int _compare(const_void *a, const_void *b):
    cdef double v = ((<Sorter*>a)).value - ((<Sorter*>b)).value
    if v < 0:
        return -1
    elif v > 0:
        return 1
    else:
        return 0


cdef int _compare_int(const_void *a, const_void *b):
    cdef int v = ((<IntSorter*>a)).value - ((<IntSorter*>b)).value
    if v < 0:
        return -1
    elif v > 0:
        return 1
    else:
        return 0


cdef void _argsort(double * data, int n_elements, Sorter * order):
    cdef int i

    for i in range(n_elements):
        order[i].index = i
        order[i].value = data[i]

    qsort(<void*> order, n_elements, sizeof(Sorter), _compare)


cdef void _argsort_int(long * data, int n_elements, IntSorter * order):
    cdef int i

    for i in range(n_elements):
        order[i].index = i
        order[i].value = data[i]

    qsort(<void*> order, n_elements, sizeof(IntSorter), _compare_int)
    # mergesort(<void*> order, n_elements, sizeof(IntSorter), _compare_int)


cdef void argsort(double * data, int n_elements, int * out):
    cdef Sorter *sorted_struct = <Sorter*>malloc(n_elements * sizeof(Sorter))

    try:
        _argsort(data, n_elements, sorted_struct)
        for i in range(n_elements):
            out[i] = sorted_struct[i].index
    finally:
        free(sorted_struct)


cdef void argsort_int(long * data, int n_elements, int * out):
    cdef IntSorter *sorted_struct = <IntSorter*>malloc(n_elements *
                                                       sizeof(IntSorter))

    try:
        _argsort_int(data, n_elements, sorted_struct)
        for i in range(n_elements):
            out[i] = sorted_struct[i].index
    finally:
        free(sorted_struct)


cdef int unique_int(long * data, int n_elements, int * out):
    cdef int * index = <int *>malloc(n_elements * sizeof(int))
    cdef int n_unique

    try:
        argsort_int(data, n_elements, index)

        if data[index[0]] != data[index[1]]:
            out[0] = data[index[0]]
        else:
            out[0] = data[index[1]]

        n_unique = 1
        for i in range(1, n_elements):
            if data[index[i]] != data[index[i - 1]]:
                out[n_unique] = data[index[i]]
                n_unique += 1
    finally:
        free(index)

    return n_unique



#! /usr/bin/env python
"""Create landlab model grids."""


from landlab.core import model_parameter_dictionary as mpd
from .raster import from_dict as raster_from_dict
from .hex import from_dict as hex_from_dict


class Error(Exception):

    """Base class for exceptions from this module."""

    pass


class BadGridTypeError(Error):

    """Raise this error for a bad grid type."""

    def __init__(self, grid_type):
        self._type = str(grid_type)

    def __str__(self):
        return self._type


_GRID_READERS = {
    'raster': raster_from_dict,
    'hex': hex_from_dict,
}


def create_and_initialize_grid(input_source):
    """Create and initialize a grid from a file.

    Creates, initializes, and returns a new grid object using parameters
    specified in *input_source*. *input_source* is either a
    ModelParameterDictionary instance (or, really, just dict-like) or a
    named input file.

    Parameters
    ----------
    input_source : str or dict
        Input file or ``dict`` of parameter values.

    Raises
    ------
    KeyError
        If missing a key from the input file.

    Examples
    --------
    >>> from six import StringIO
    >>> test_file = StringIO('''
    ... GRID_TYPE:
    ... raster
    ... NUM_ROWS:
    ... 4
    ... NUM_COLS:
    ... 5
    ... GRID_SPACING:
    ... 2.5
    ... ''')
    >>> from landlab import create_and_initialize_grid
    >>> grid = create_and_initialize_grid(test_file)
    >>> grid.number_of_nodes
    20
    """
    if isinstance(input_source, dict):
        param_dict = input_source
    else:
        param_dict = mpd.ModelParameterDictionary(from_file=input_source)

    grid_type = param_dict['GRID_TYPE']

    grid_type.strip().lower()

    # Read parameters appropriate to that type, create it, and initialize it
    try:
        grid_reader = _GRID_READERS[grid_type]
    except KeyError:
        raise BadGridTypeError(grid_type)

    # Return the created and initialized grid
    return grid_reader(param_dict)



"""Landlab component that generates a random fire event in time.

This component generates a random fire event or fire time series from the
Weibull statistical distribution.

.. codeauthor:: Jordan Adams

This component generates random numbers using the Weibull distribution
(Weibull, 1951). No particular units must be used, but it was written with
the fire recurrence units in time (yrs).

Using the Weibull Distribution assumes two things: All elements within the
study area have the same fire regime. Each element must have (on average) a
constant fire regime during the time span of the study.

As of Sept. 2013, fires are considered instantaneous events independent of
other fire events in the time series.

Written by Jordan M. Adams, 2013. Updated April 2016.


Examples
--------
>>> from landlab.components.fire_generator import FireGenerator

Create an instance of the FireGenerator component


>>> fg = FireGenerator(mean_fire_recurrence = 15.0, shape_parameter = 4.5)

This creates an instance of the component that has a mean_fire_recurrence, or
average interval between fires of 15 years. We gave it a shape parameter of
4.5, suggesting the cumulative distribution function that is skewed right.

Since we didn't pass it a scale parameter, the component calculates it for you.
Testing this...

>>> fg.scale_parameter
16.437036931437866

To get a time to next fire:

>>> fg.generate_fire_recurrence()  # doctest: +SKIP
10.68

"""

from random import weibullvariate
from scipy import special
from landlab import Component


class FireGenerator(Component):

    """Generate a random fire event or time series.

    Parameters
    ----------
    mean_fire_recurrence : float
        Average time between fires for a given location
    shape_parameter : float
        Describes the skew of the Weibull distribution.
        If shape < 3.5, data skews left.
        If shape == 3.5, data is normal.
        If shape > 3.5, data skews right.
        To approximate a normal bell curve, use a value of 3.5
    scale_parameter : float, optional
        Describes the peak of the Weibull distribution, located at the
        63.5% value of the cumulative distribution function. If unknown,
        it can be found using mean fire recurrence value and the
        get_scale_parameter() method described later.
    """

    _name = 'FireGenerator'

    _input_var_names = tuple()

    _output_var_names = tuple()

    _var_units = dict()

    _var_mapping = dict()

    _var_doc = dict()

    def __init__(self, mean_fire_recurrence=0.0, shape_parameter=0.0,
                 scale_parameter=None):
        """Generate a random fire event in time.

        Parameters
        ----------
        mean_fire_recurrence : float
            Average time between fires for a given location
        shape_parameter : float
            Describes the skew of the Weibull distribution.
            If shape < 3.5, data skews left.
            If shape == 3.5, data is normal.
            If shape > 3.5, data skews right.
        scale_parameter : float, optional
            Describes the peak of the Weibull distribution, located at the
            63.5% value of the cumulative distribution function. If unknown,
            it can be found using mean fire recurrence value and the
            get_scale_parameter().
        """

        self.mean_fire_recurrence = mean_fire_recurrence

        self.shape_parameter = shape_parameter

        if scale_parameter is None:
            self.get_scale_parameter()

        else:
            self.scale_parameter = scale_parameter


    def get_scale_parameter(self):
        """Get the scale parameter.

        ::
            mean_fire_recurrence = (scale_parameter * (
                                    special.gamma(1 + (1 / shape))))

        sets the scale parameter.
        """

        shape_in_gamma_func = float( 1+ (1 / self.shape_parameter))
        gamma_func = special.gamma(shape_in_gamma_func)
        self.scale_parameter = (self.mean_fire_recurrence / gamma_func)


    def generate_fire_recurrence(self):
        """Get time to next fire.

        Finds the time to next fire (fire recurrence) based on the scale
        parameter (63.5% of fire Weibull distribution) and the shape parameter
        (describes the skew of the histogram, shape = 3.5
        represents a normal distribution).

        Rounds the time to next fire to 4 significant figures, for neatness.

        Returns
        -------
        float
            Updated value for the time to next fire.

        """
        self.time_to_next_fire = round(weibullvariate(self.scale_parameter,
                                                      self.shape_parameter), 2)
        return self.time_to_next_fire



import numpy as np
cimport numpy as np
cimport cython
from libc.stdlib cimport malloc, free


DTYPE = np.int
ctypedef np.int_t DTYPE_t


@cython.boundscheck(False)
def remove_patches(np.ndarray[DTYPE_t, ndim=2] links_at_patch,
                   np.ndarray[DTYPE_t, ndim=1] patches_to_remove):
    cdef int n_bad_patches = len(patches_to_remove)
    cdef int n_patches = links_at_patch.shape[0]
    cdef int max_links = links_at_patch.shape[1]
    cdef int patch
    cdef int n

    new_patch = 0
    for patch in range(n_patches):
        if patch in patches_to_remove:
            pass
        else:
            for n in range(max_links):
                links_at_patch[new_patch, n] = links_at_patch[patch, n]
            patch += 1


@cython.boundscheck(False)
def remove_tris(np.ndarray[DTYPE_t, ndim=2] nodes_at_tri,
                np.ndarray[DTYPE_t, ndim=2] neighbors_at_tri,
                np.ndarray[DTYPE_t, ndim=1] bad_tris):
    cdef int n_tris = nodes_at_tri.shape[0]
    cdef int n_bad_tris = len(bad_tris)
    cdef int n_patches = n_tris - n_bad_tris
    cdef int tri
    cdef int patch
    cdef int *patch_at_tri = <int *>malloc(n_tris * sizeof(int))
    cdef int n
    cdef int old

    try:
        patch = 0
        for tri in range(n_tris):
            if tri in bad_tris:
                patch_at_tri[tri] = -1
            else:
                for n in range(3):
                  nodes_at_tri[patch, n] = nodes_at_tri[tri, n]
                  neighbors_at_tri[patch, n] = neighbors_at_tri[tri, n]
                patch_at_tri[tri] = patch
                patch += 1

        for tri in range(n_tris):
            for n in range(3):
                old = neighbors_at_tri[tri, n]
                if old >= 0:
                    neighbors_at_tri[tri, n] = patch_at_tri[old]
                else:
                    neighbors_at_tri[tri, n] = -1
    finally:
        free(patch_at_tri)


@cython.boundscheck(False)
def _setup_links_at_patch(np.ndarray[DTYPE_t, ndim=2] nodes_at_patch,
                          np.ndarray[DTYPE_t, ndim=2] tri_neighbors,
                          np.ndarray[DTYPE_t, ndim=2] nodes_at_link,
                          np.ndarray[DTYPE_t, ndim=2] links_at_patch):
  cdef int i
  cdef int link
  cdef int neighbor
  cdef int n_patches = len(nodes_at_patch)
  cdef int *tri_done = <int *>malloc(n_patches * sizeof(int))
  cdef int *links_per_patch = <int *>malloc(n_patches * sizeof(int))

  if not tri_done or not links_per_patch:
    raise MemoryError(
      'unable to allocate {bytes} bytes'.format(bytes=n_patches * sizeof(int)))

  try:
    for tri in range(n_patches):
      tri_done[tri] = 0
      links_per_patch[tri] = 0

    link = 0
    for tri in range(n_patches):
      for i in (0, 1, 2):
        neighbor = tri_neighbors[tri, i]

        if neighbor == -1 or not tri_done[neighbor]:
          nodes_at_link[link, 0] = nodes_at_patch[tri, (i + 1) % 3]
          nodes_at_link[link, 1] = nodes_at_patch[tri, (i + 2) % 3]

          links_at_patch[tri, links_per_patch[tri]] = link
          links_per_patch[tri] += 1

          if neighbor >= 0:
            links_at_patch[neighbor, links_per_patch[neighbor]] = link
            links_per_patch[neighbor] += 1
            tri_done[tri] = True

          link += 1
  finally:
    free(links_per_patch)
    free(tri_done)



#! /usr/bin/env python
import os

import numpy as np
from numpy.testing import assert_array_almost_equal

from landlab.io import read_esri_ascii
from landlab.io.netcdf import read_netcdf
from landlab import RasterModelGrid


def test_save_esri_ascii(tmpdir):
    grid = RasterModelGrid(4, 5, 2.)
    grid.add_field('node', 'air__temperature', np.arange(20.))

    with tmpdir.as_cwd():
        grid.save('test.asc', format='esri-ascii')
        assert os.path.isfile('test.asc')


def test_add_extension(tmpdir):
    grid = RasterModelGrid(4, 5, 2.)
    grid.add_field('node', 'air__temperature', np.arange(20.))

    with tmpdir.as_cwd():
        grid.save('test', format='esri-ascii')
        assert os.path.isfile('test.asc')

    with tmpdir.as_cwd():
        grid.save('test', format='netcdf')
        assert os.path.isfile('test.nc')


def test_replace_extension(tmpdir):
    grid = RasterModelGrid(4, 5, 2.)
    grid.add_field('node', 'air__temperature', np.arange(20.))

    with tmpdir.as_cwd():
        grid.save('test.nc', format='esri-ascii')
        assert os.path.isfile('test.asc')

    with tmpdir.as_cwd():
        grid.save('test.asc', format='netcdf')
        assert os.path.isfile('test.nc')


def test_guess_format(tmpdir):
    grid = RasterModelGrid(4, 5, 2.)
    grid.add_field('node', 'air__temperature', np.arange(20.))

    with tmpdir.as_cwd():
        grid.save('test.asc')
        assert os.path.isfile('test.asc')
        read_esri_ascii('test.asc')

    with tmpdir.as_cwd():
        grid.save('test.nc')
        assert os.path.isfile('test.nc')
        read_netcdf('test.nc')


def test_names_keyword_as_str(tmpdir):
    grid = RasterModelGrid(4, 5, 2.)
    grid.add_field('air__temperature', np.arange(20.), at='node')
    grid.add_field('land_surface__elevation', np.arange(20.), at='node')

    with tmpdir.as_cwd():
        grid.save('test.asc', names='land_surface__elevation')
        assert os.path.isfile('test.asc')
        read_esri_ascii('test.asc')


def test_names_keyword_as_list(tmpdir):
    grid = RasterModelGrid(4, 5, 2.)
    grid.add_field('air__temperature', np.arange(20.), at='node')
    grid.add_field('land_surface__elevation', np.arange(20.), at='node')

    with tmpdir.as_cwd():
        grid.save('test.asc', names=['land_surface__elevation',
                                     'air__temperature'])
        files = ['test_land_surface__elevation.asc',
                 'test_air__temperature.asc']
        for fname in files:
            assert os.path.isfile(fname)
            read_esri_ascii(fname)



#! /usr/bin/env python
"""Return array with same shape as grid elements."""

import numpy as np

from landlab.utils.decorators import use_field_name_array_or_value

@use_field_name_array_or_value('node')
def return_array_at_node(grid, value):
    """Function to return an array stored at node or of shape `(n_nodes,)`.

    This function exists to take advantange of the use_field_name_array_or_value
    decorator which permits providing the surface as a field name or array.

    Parameters
    ----------
    grid : ModelGrid
    value : field name, ndarray of shape `(n_nodes, )`, or single value.

    Returns
    -------
    array : ndarray of shape `(n_nodes, )`
    """
    return value



"""Test the flow director components.

@author: krb
"""
# Created on Thurs Nov 12, 2015
import os

import pytest
from six.moves import range
import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal

import landlab
from landlab import RasterModelGrid, HexModelGrid, RadialModelGrid, FieldError
from landlab.components.flow_director import(FlowDirectorD8,
                                             FlowDirectorDINF,
                                             FlowDirectorMFD,
                                             FlowDirectorSteepest)

from landlab.components.flow_director.flow_director import _FlowDirector
from landlab.components.flow_director.flow_director_to_many import _FlowDirectorToMany
from landlab.components.flow_director.flow_director_to_one import _FlowDirectorToOne

from landlab import CLOSED_BOUNDARY
from landlab import BAD_INDEX_VALUE as XX


_THIS_DIR = os.path.abspath(os.path.dirname(__file__))

def test_not_implemented():
    """Test that private run_one_step is not implemented"""

    mg = RasterModelGrid((10,10), spacing=(1, 1))
    z = mg.add_field('topographic__elevation', mg.node_x**2 + mg.node_y**2, at = 'node')

    fd0 = _FlowDirector(mg, 'topographic__elevation')
    with pytest.raises(NotImplementedError):
        fd0.run_one_step()

    fd1 = _FlowDirectorToMany(mg, 'topographic__elevation')
    with pytest.raises(NotImplementedError):
        fd1.run_one_step()

    fd2 = _FlowDirectorToOne(mg, 'topographic__elevation')
    with pytest.raises(NotImplementedError):
        fd2.run_one_step()


def test_grid_type_testing():
    """Test that only the right grids can be implemented."""
    dx=(2./(3.**0.5))**0.5
    hmg = HexModelGrid(9,5, dx)
    z = hmg.add_field('topographic__elevation', hmg.node_x + np.round(hmg.node_y), at = 'node')

    # D8 is ONLY RASTER
    with pytest.raises(NotImplementedError):
        FlowDirectorD8(hmg)

    # DINF IS ONLY RASTER RASTER
    with pytest.raises(NotImplementedError):
        FlowDirectorDINF(hmg)

def test_check_fields():
    """Check to make sure the right fields have been created.

    Check that the sizes of at least one are also correct (they are at node
    so if one is the right size, then they all should be)
    """

    mg0 = RasterModelGrid((10,10), spacing=(1, 1))
    z0 = mg0.add_field('topographic__elevation', mg0.node_x**2 + mg0.node_y**2, at = 'node')
    fd0 = _FlowDirector(mg0, 'topographic__elevation')
    assert list(mg0.at_node.keys()) == ['topographic__elevation']
    assert np.size(mg0.at_node['topographic__elevation']) == mg0.number_of_nodes

    mg1 = RasterModelGrid((10,10), spacing=(1, 1))
    z1 = mg1.add_field('topographic__elevation', mg1.node_x**2 + mg1.node_y**2, at = 'node')
    fd1 = _FlowDirectorToMany(mg1, 'topographic__elevation')
    assert sorted(list(mg1.at_node.keys())) == [
        'flow__link_to_receiver_node',
        'flow__receiver_node',
        'flow__sink_flag',
        'topographic__elevation',
        'topographic__steepest_slope',
    ]
    assert np.size(mg1.at_node['topographic__elevation']) == mg1.number_of_nodes

    mg2 = RasterModelGrid((10,10), spacing=(1, 1))
    z2 = mg2.add_field('topographic__elevation', mg2.node_x**2 + mg2.node_y**2, at = 'node')
    fd2 = _FlowDirectorToOne(mg2, 'topographic__elevation')
    assert sorted(list(mg2.at_node.keys())) == [
        'flow__link_to_receiver_node',
        'flow__receiver_node',
        'flow__sink_flag',
        'topographic__elevation',
        'topographic__steepest_slope',
    ]
    assert np.size(mg2.at_node['topographic__elevation']) == mg2.number_of_nodes


    mg3 = RasterModelGrid((10,10), spacing=(1, 1))
    z3 = mg3.add_field('topographic__elevation', mg3.node_x**2 + mg3.node_y**2, at = 'node')
    fd3 = FlowDirectorMFD(mg3, 'topographic__elevation')
    assert sorted(list(mg3.at_node.keys())) == [
        'flow__link_to_receiver_node',
        'flow__links_to_receiver_nodes',
        'flow__receiver_node',
        'flow__receiver_nodes',
        'flow__receiver_proportions',
        'flow__sink_flag',
        'topographic__elevation',
        'topographic__steepest_slope',
    ]
    assert np.size(mg3.at_node['topographic__elevation']) == mg3.number_of_nodes

    mg4 = RasterModelGrid((10,10), spacing=(1, 1))
    z4 = mg4.add_field('topographic__elevation', mg4.node_x**2 + mg4.node_y**2, at = 'node')
    fd4 = FlowDirectorDINF(mg4, 'topographic__elevation')
    assert sorted(list(mg4.at_node.keys())) == [
        'flow__link_to_receiver_node',
        'flow__links_to_receiver_nodes',
        'flow__receiver_node',
        'flow__receiver_nodes',
        'flow__receiver_proportions',
        'flow__sink_flag',
        'topographic__elevation',
        'topographic__steepest_slope',
    ]
    assert np.size(mg4.at_node['topographic__elevation']) == mg4.number_of_nodes

    mg5 = RasterModelGrid((10,10), spacing=(1, 1))
    z5 = mg5.add_field('topographic__elevation', mg5.node_x**2 + mg5.node_y**2, at = 'node')
    fd5 = FlowDirectorSteepest(mg5, 'topographic__elevation')
    assert sorted(list(mg5.at_node.keys())) == [
        'flow__link_to_receiver_node',
        'flow__receiver_node',
        'flow__sink_flag',
        'topographic__elevation',
        'topographic__steepest_slope',
    ]
    assert np.size(mg5.at_node['topographic__elevation']) == mg5.number_of_nodes

    mg6 = RasterModelGrid((10,10), spacing=(1, 1))
    z6 = mg6.add_field('topographic__elevation', mg6.node_x**2 + mg6.node_y**2, at = 'node')
    fd6 = FlowDirectorD8(mg6, 'topographic__elevation')
    assert sorted(list(mg6.at_node.keys())) == [
        'flow__link_to_receiver_node',
        'flow__receiver_node',
        'flow__sink_flag',
        'topographic__elevation',
        'topographic__steepest_slope',
    ]
    assert np.size(mg6.at_node['topographic__elevation']) == mg6.number_of_nodes



import numpy as np
cimport numpy as np
cimport cython

from libc.stdlib cimport malloc, free

DTYPE = np.int
ctypedef np.int_t DTYPE_t


@cython.boundscheck(False)
def calc_area_at_patch(np.ndarray[long, ndim=2] nodes_at_patch,
                       np.ndarray[np.float_t, ndim=1] x_of_node,
                       np.ndarray[np.float_t, ndim=1] y_of_node,
                       np.ndarray[np.float_t, ndim=1] out):
    cdef long n_patches = nodes_at_patch.shape[0]
    cdef long n_vertices = nodes_at_patch.shape[1]
    cdef long n

    for n in range(n_patches):
      out[n] = calc_area_of_patch(&nodes_at_patch[n, 0], n_vertices,
                                  &x_of_node[0], &y_of_node[0])


cdef calc_area_of_patch(long * nodes_at_patch, long n_vertices,
                        double * x_of_node, double * y_of_node):
    cdef int n
    cdef int node
    cdef float * x_of_vertex = <float *>malloc(n_vertices * sizeof(float))
    cdef float * y_of_vertex = <float *>malloc(n_vertices * sizeof(float))

    try:
        for n in range(n_vertices):
            node = nodes_at_patch[n]
            if node == -1:
                n -= 1
                break
            x_of_vertex[n] = x_of_node[node]
            y_of_vertex[n] = y_of_node[node]

        return calc_area_of_polygon(x_of_vertex, y_of_vertex, n + 1)
    finally:
        free(y_of_vertex)
        free(x_of_vertex)


@cython.boundscheck(False)
def calc_centroid_at_patch(np.ndarray[long, ndim=2] nodes_at_patch,
                           np.ndarray[np.float_t, ndim=1] x_of_node,
                           np.ndarray[np.float_t, ndim=1] y_of_node,
                           np.ndarray[np.float_t, ndim=2] out):
    cdef long n_patches = nodes_at_patch.shape[0]
    cdef long n_vertices = nodes_at_patch.shape[1]
    cdef long n

    for n in range(n_patches):
        calc_centroid_of_patch(&nodes_at_patch[n, 0], n_vertices,
                               &x_of_node[0], &y_of_node[0],
                               &out[n, 0])


cdef calc_centroid_of_patch(long * nodes_at_patch, long n_vertices,
                            double * x_of_node, double * y_of_node, double * out):
    cdef int n
    cdef int node
    cdef float * x = <float *>malloc(n_vertices * sizeof(float))
    cdef float * y = <float *>malloc(n_vertices * sizeof(float))

    try:
        for n in range(n_vertices):
            node = nodes_at_patch[n]
            if node == -1:
                n -= 1
                break
            x[n] = x_of_node[node]
            y[n] = y_of_node[node]
        calc_centroid_of_polygon(x, y, n + 1, out)
    finally:
        free(y)
        free(x)


cdef calc_centroid_of_polygon(float * x, float * y, long n_vertices,
                              double * out):
    cdef float x_of_centroid = 0.
    cdef float y_of_centroid = 0.
    cdef float area = calc_area_of_polygon(x, y, n_vertices)
    cdef float c
    cdef int n

    c = x[n_vertices - 1] * y[0] - x[0] * y[n_vertices - 1]
    x_of_centroid = (x[n_vertices - 1] + x[0]) * c
    y_of_centroid = (y[n_vertices - 1] + y[0]) * c

    for n in range(n_vertices - 1):
        c = x[n] * y[n + 1] - x[n + 1] * y[n]
        x_of_centroid += (x[n] + x[n + 1]) * c
        y_of_centroid += (y[n] + y[n + 1]) * c

    # c = x[n_vertices - 1] * y[0] - x[0] - y[n_vertices - 1]
    # x_of_centroid += (x[n_vertices - 1] + x[0]) * c
    # y_of_centroid += (y[n_vertices - 1] + y[0]) * c

    x_of_centroid /= 6. * area
    y_of_centroid /= 6. * area

    out[0] = x_of_centroid
    out[1] = y_of_centroid


cdef calc_area_of_polygon(float * x, float * y, long n_vertices):
    cdef float area = 0.
    cdef int n

    for n in range(n_vertices - 1):
        area += x[n] * y[n + 1]
        area -= x[n + 1] * y[n]
    area += x[n_vertices - 1] * y[0]
    area -= x[0] * y[n_vertices - 1]
    area /= 2.

    return area



from .sort import (sort_graph, sort_patches, sort_nodes, reindex_by_xy,
                   reorder_links_at_patch)


__all__ = ['sort_graph', 'sort_patches', 'sort_nodes', 'reindex_by_xy',
           'reorder_links_at_patch', ]



#!/usr/env/python

"""
Short description.

flow_accum_to_n.py: Implementation a route-to-multiple drainage stack alorithm.


Algorithm for route to multiple (N) flow accumulation. Inspiration for data
structures and attempting O(n) efficiency taken from Braun and Willet(2013).

Algorithm constructs drainage area and (optionally) water discharge. Can
handle the case in which each node has more than one downstream receiver.

Computationally, for a grid of the same size this algorithm will take about
    1.5*(avg number of downstream nodes per cell)
        *(duration of flow_accum_bw for same grid using route-to-one method)

So under route-to-one direction schemes, using the Braun and Willet method is
recommended.

If water discharge is calculated, the result assumes steady flow (that is,
hydrologic equilibrium).

The main public function is:

    a, q, s = flow_accumulation_to_n(r, p)

which takes the following inputs:

    r, an (np, q) array of receiver-node IDs, where np is the total number of
    nodes and q is the maximum number of receivers any node in the grid has.
    This array would be returned by the flow_routing component.

    p, an (np, q) array that identifies the proportion of flow going to each
    receiver. For each q elements along the np axis, sum(p(i, :)) must equal
    1. This array would be returned by the flow_routing component.

It returns Numpy arrays with the drainage area (a) and discharge (q) at each
node, along with an array (s) that contains the IDs of the nodes in downstream-
to-upstream order.

If you simply want the ordered list by itself, use::

    s = make_ordered_node_array_to_n(r, p, b)

Created: KRB Oct 2016 (modified from flow_accumu_bw)
"""
import numpy
from six.moves import range

class _DrainageStack_to_n():

    """
    Implementation of the DrainageStack_to_n class.

    The _DrainageStack_to_n() class implements a set based approach to
    constructing a stack with similar properties to the stack constructed by
    Braun & Willet (2013). It constructs an list, s, of all nodes in the grid
    such that a given node is always located earlier in the list than all
    upstream nodes that contribute to it.

    It is used by the make_ordered_node_array_to_n() function.
    """


    def __init__(self, delta, D, num_receivers):

        """
        Creates the stack array s and stores references to delta and D.

        Initialization of the _DrainageStack_to_n() class including storing
        delta and D.
        """
        
        self.num_receivers = num_receivers
        self.s = list()
        self.delta = delta
        self.D = D


    def construct__stack(self, l):
        """
        Function to construct the drainage stack.

        Function to add all nodes upstream of a set of base level nodes given
        by list l in an order
        such that downstream nodes always occur before upstream nodes.

        This function contains the major algorithmic difference between the
        route to 1 method of Braun and Willet (2013) and the route to N method
        presented here.

        Rather than recursively moving up the tributary tree this method uses
        sets test that a node is downstream and add it to the stack. Both
        methods are functionally depth first searches. The method that Braun
        and Willet (2013) implement is optimized given that each node only has
        one receiver. This method is optimized to visit more than one vertex/
        node of the graph at a time.

        An important note: Since sets are un-ordered, we cannot expect the
        stack to be exactly the same each time. It will always put nodes that
        are downstream before those that are upstream, but because it will move
        up multiple branches at the same time, it may put three nodes into the
        stack at the same time that are on different branches of the flow
        network. Because these nodes are in different parts of the network,
        the relative order of them does not matter.

        For example, in the example below, the nodes 1 and 7 must be added
        after 5 but before 2 and 6.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab.components.flow_accum.flow_accum_to_n import(
        ... _DrainageStack_to_n)
        >>> delta = np.array([0, 0, 2, 4, 4, 8, 12, 14, 17, 18, 18])
        >>> num_receivers = np.array([2, 2, 2, 2, 1, 1, 2, 2, 2, 2])
        >>> D = np.array([0, 2, 0, 3, 1, 4, 5, 7, 6, 1, 2, 7, 3, 8, 9, 6, 8, 9])
        >>> ds = _DrainageStack_to_n(delta, D, num_receivers)
        >>> ds.construct__stack(4)
        >>> ds.s[0] == 4
        True
        >>> ds.s[1] == 5
        True
        >>> ds.s[9] == 9
        True
        >>> len(set([1, 7])-set(ds.s[2:4]))
        0
        >>> len(set([2, 6])-set(ds.s[4:6]))
        0
        >>> len(set([0, 3, 8])-set(ds.s[6:9]))
        0
        """
        # create base nodes set
        try:
            base = set(l)
        except:
            base = set([l])
        
        # instantiate the time keeping variable i, and a variable to keep track
        # of the visit time. Using visit time allows us to itterate through
        # the entire graph and make sure that only put a node in the stack 
        # the last time it is visited. 
        
        i = 0
        visit_time = -1*numpy.ones((self.delta.size-1))
        num_visits = numpy.zeros((self.delta.size-1))
        
        # deal with the first node, which goes to it 
        visit_time[list(base)] = i
        num_visits[list(base)] += 1 
        
        i = 1
        visited = set([])
        for node_i in base:
                # select the nodes to visit
                visit = set(self.D[self.delta[node_i]:self.delta[node_i+1]])
                visit = visit-base
                
                # record the visit time.
                visit_time[list(visit)] = i
                          
                # record that they have been visited. 
                num_visits[list(visit)] += 1
                
                visited.update(list(visit))
        
        visited = numpy.array(list(visited))
        visited_enough = num_visits[visited]==self.num_receivers[visited]
        completed = set(visited[visited_enough])
        
        # recurse through the remainder. Only look above completed nodes, 
        # this prevents repeat link walking. 
        while len(completed)>0:
            # increase counter
            i += 1
            
            visited = set([])
            new_completes = set([])
            
            for node_i in completed:
                    
                    # select the nodes to visit
                    visit = self.D[self.delta[node_i]:self.delta[node_i+1]]
                    # record the visit time.
                    visit_time[visit] = i
                              
                    # record that they have been visited. 
                    num_visits[visit] += 1
                              
                    # add nodes that have been visited enough times to complete
                    # to the upstream stack. We can ignore the rest, they will
                    # be re-visited. This should reduce the number of times each
                    # link is walked to the number of active links. 
                    visited_enough = num_visits[numpy.array(visit)]==self.num_receivers[numpy.array(visit)]
                    
                    visited.update(visit)
                    new_completes.update(visit[visited_enough])
            completed = new_completes
            
        
        
        # the stack is the argsort of visit time. 
        self.s = numpy.argsort(visit_time)

def _make_number_of_donors_array_to_n(r, p):

    """Number of donors for each node.

    Creates and returns an array containing the number of donors for each node.

    Parameters
    ----------
    r : ndarray size (np, q) where r[i,:] gives all receivers of node i. Each
        node recieves flow fom up to q donors.

    p : ndarray size (np, q) where p[i,v] give the proportion of flow going
        from node i to the receiver listed in r[i,v].

    Returns
    -------
    ndarray size (np)
        Number of donors for each node.

    Examples
    --------

    >>> import numpy as np
    >>> from landlab.components.flow_accum.flow_accum_to_n import(
    ... _make_number_of_donors_array_to_n)
    >>> r = np.array([[ 1,  2],
    ...               [ 4,  5],
    ...               [ 1,  5],
    ...               [ 6,  2],
    ...               [ 4, -1],
    ...               [ 4, -1],
    ...               [ 5,  7],
    ...               [ 4,  5],
    ...               [ 6,  7],
    ...               [ 7,  8]])
    >>> p = np.array([[ 0.6,   0.4 ],
    ...               [ 0.85,  0.15],
    ...               [ 0.65,  0.35],
    ...               [ 0.9,   0.1 ],
    ...               [ 1.,    0.  ],
    ...               [ 1.,    0.  ],
    ...               [ 0.75,  0.25],
    ...               [ 0.55,  0.45],
    ...               [ 0.8,   0.2 ],
    ...               [ 0.95,  0.05]])
    >>> nd = _make_number_of_donors_array_to_n(r, p)
    >>> nd
    array([0, 2, 2, 0, 4, 4, 2, 3, 1, 0])
    """

    # Vectorized, DEJH, 5/20/14
#    np = len(r)
#    nd = numpy.zeros(np, dtype=int)
#    for i in range(np):
#        nd[r[i]] += 1

    # modified by KRB 10/31/2016 to support route to multiple.

    nd = numpy.zeros(r.shape[0], dtype=int)
    max_index = numpy.amax(r)

    # filter r based on p and flatten
    r_filter_flat = r.flatten()[p.flatten() > 0]

    nd[:(max_index + 1)] = numpy.bincount(r_filter_flat)
    return nd


def _make_delta_array_to_n(nd):
    r"""
    Function to create the delta array.

    Creates and returns the "delta" array, which is a list containing, for each
    node, the array index where that node's donor list begins.

    Parameters
    ----------
    nd : ndarray of int
        Number of donors for each node

    Returns
    -------
    ndarray of int
        Delta array

    Examples
    --------

    >>> import numpy as np
    >>> from landlab.components.flow_accum.flow_accum_to_n import(
    ... _make_delta_array_to_n)
    >>> nd = np.array([0, 2, 2, 0, 4, 4, 2, 3, 1, 0])
    >>> delta = _make_delta_array_to_n(nd)
    >>> delta
    array([ 0,  0,  2,  4,  4,  8,  12,  14, 17, 18, 18])
    >>> sum(nd) == max(delta)
    True
    """
    # np = len(nd)
    # delta = numpy.zeros(np+1, dtype=int)
    # delta[np] = np   # not np+1 as in B&W because here we number from 0
    # for i in range(np-1, -1, -1):
    #     delta[i] = delta[i+1] - nd[i]
    # return delta

    # DEJH efficient delooping (only a small gain)

    nt = sum(nd)
    np = len(nd)
    delta = numpy.zeros(np+1, dtype=int)
    delta.fill(nt)
    delta[-2::-1] -= numpy.cumsum(nd[::-1])

    return delta


def _make_array_of_donors_to_n(r, p, delta):

    """
    Creates and returns an array containing the IDs of donors for each node.

    Essentially, the array is a series of lists (not in the Python list object
    sense) of IDs for each node. See Braun & Willett (2012) for details.

    The example below is from Braun & Willett (2012), and produces D_i in their
    Table 1 (except that here the ID numbers are one less, because we number
    indices from zero).

    Vectorized - inefficiently! - DEJH, 5/20/14

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.components.flow_accum.flow_accum_to_n import(
    ... _make_array_of_donors_to_n)
    >>> r = np.array([[ 1,  2],
    ...               [ 4,  5],
    ...               [ 1,  5],
    ...               [ 6,  2],
    ...               [ 4, -1],
    ...               [ 4, -1],
    ...               [ 5,  7],
    ...               [ 4,  5],
    ...               [ 6,  7],
    ...               [ 7,  8]])
    >>> p = np.array([[ 0.6,   0.4 ],
    ...               [ 0.85,  0.15],
    ...               [ 0.65,  0.35],
    ...               [ 0.9,   0.1 ],
    ...               [ 1.,    0.  ],
    ...               [ 1.,    0.  ],
    ...               [ 0.75,  0.25],
    ...               [ 0.55,  0.45],
    ...               [ 0.8,   0.2 ],
    ...               [ 0.95,  0.05]])
    >>> delta = np.array([ 0,  0,  2,  4,  4,  8,  12,  14, 17, 18, 18])
    >>> D = _make_array_of_donors_to_n(r, p, delta)
    >>> D
    array([0, 2, 0, 3, 1, 4, 5, 7, 6, 1, 2, 7, 3, 8, 9, 6, 8, 9])
    """
    np = r.shape[0]
    q = r.shape[1]
    nt = delta[-1]

    w = numpy.zeros(np, dtype=int)
    D = numpy.zeros(nt, dtype=int)
    for v in range(q):
        for i in range(np):
            ri = r[i, v]
            if p[i, v] > 0:
                ind = delta[ri]+w[ri]
                D[ind] = i
                w[ri] += 1

    return D

    # DEJH notes that for reasons he's not clear on, this looped version is
    # actually much slower!
    # D = numpy.zeros(np, dtype=int)
    # wri_fin = numpy.bincount(r)
    # wri_fin_nz = wri_fin.nonzero()[0]
    # wri_fin_nz_T = wri_fin_nz.reshape((wri_fin_nz.size,1))
    # logical = numpy.tile(r,(wri_fin_nz.size,1))==wri_fin_nz_T
    # cum_logical = numpy.cumsum(logical, axis=1)
    # wri = numpy.sum(numpy.where(logical, cum_logical-1,0) ,axis=0)
    # D_index = delta[r] + wri
    # D[D_index] = numpy.arange(r.size)
    # return D


def make_ordered_node_array_to_n(receiver_nodes,
                                 receiver_proportion):

    """Create an array of node IDs.

    Creates and returns an array of node IDs that is arranged in order from
    downstream to upstream.

    The lack of a leading underscore is meant to signal that this operation
    could be useful outside of this module!

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.components.flow_accum.flow_accum_to_n import(
    ... make_ordered_node_array_to_n)
    >>> r = np.array([[ 1,  2],
    ...               [ 4,  5],
    ...               [ 1,  5],
    ...               [ 6,  2],
    ...               [ 4, -1],
    ...               [ 4, -1],
    ...               [ 5,  7],
    ...               [ 4,  5],
    ...               [ 6,  7],
    ...               [ 7,  8]])
    >>> p = np.array([[ 0.6,   0.4 ],
    ...               [ 0.85,  0.15],
    ...               [ 0.65,  0.35],
    ...               [ 0.9,   0.1 ],
    ...               [ 1.,    0.  ],
    ...               [ 1.,    0.  ],
    ...               [ 0.75,  0.25],
    ...               [ 0.55,  0.45],
    ...               [ 0.8,   0.2 ],
    ...               [ 0.95,  0.05]])
    >>> s = make_ordered_node_array_to_n(r, p)
    >>> s[0] == 4
    True
    >>> s[1] == 5
    True
    >>> s[9] == 9
    True
    >>> len(set([1, 7])-set(s[2:4]))
    0
    >>> len(set([2, 6])-set(s[4:6]))
    0
    >>> len(set([0, 3, 8])-set(s[6:9]))
    0
    """
    node_id = numpy.arange(receiver_nodes.shape[0])
    baselevel_nodes = numpy.where(node_id==receiver_nodes[:,0])[0]
    nd = _make_number_of_donors_array_to_n(receiver_nodes, receiver_proportion)
    delta = _make_delta_array_to_n(nd)
    D = _make_array_of_donors_to_n(receiver_nodes, receiver_proportion, delta)
    
    num_receivers = numpy.sum(receiver_nodes>=0, axis=1)
    
    dstack = _DrainageStack_to_n(delta, D, num_receivers)
    construct_it = dstack.construct__stack

    construct_it(baselevel_nodes)  # don't think this is a bottleneck, so no C++
    return dstack.s



def find_drainage_area_and_discharge_to_n(s, r, p, node_cell_area=1.0,
                                          runoff=1.0, boundary_nodes=None):

    """Calculate the drainage area and water discharge at each node.

    Parameters
    ----------
    s : ndarray of int
        Ordered (downstream to upstream) array of node IDs
    r : ndarray size (np, q) where r[i, :] gives all receivers of node i. Each
        node recieves flow fom up to q donors.
    p : ndarray size (np, q) where p[i, v] give the proportion of flow going
        from node i to the receiver listed in r[i, v].
    node_cell_area : float or ndarray
        Cell surface areas for each node. If it's an array, must have same
        length as s (that is, the number of nodes).
    runoff : float or ndarray
        Local runoff rate at each cell (in water depth per time). If it's an
        array, must have same length as s (that is, the number of nodes).

    Returns
    -------
    tuple of ndarray
        drainage area and discharge

    Notes
    -----
    -  If node_cell_area not given, the output drainage area is equivalent
       to the number of nodes/cells draining through each point, including
       the local node itself.
    -  Give node_cell_area as a scalar when using a regular raster grid.
    -  If runoff is not given, the discharge returned will be the same as
       drainage area (i.e., drainage area times unit runoff rate).
    -  If using an unstructured Landlab grid, make sure that the input
       argument for node_cell_area is the cell area at each NODE rather than
       just at each CELL. This means you need to include entries for the
       perimeter nodes too. They can be zeros.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.components.flow_accum.flow_accum_to_n import(
    ... find_drainage_area_and_discharge_to_n)
    >>> r = np.array([[ 1,  2],
    ...               [ 4,  5],
    ...               [ 1,  5],
    ...               [ 6,  2],
    ...               [ 4, -1],
    ...               [ 4, -1],
    ...               [ 5,  7],
    ...               [ 4,  5],
    ...               [ 6,  7],
    ...               [ 7,  8]])
    >>> p = np.array([[ 0.6,   0.4 ],
    ...               [ 0.85,  0.15],
    ...               [ 0.65,  0.35],
    ...               [ 0.9,   0.1 ],
    ...               [ 1.,    0.  ],
    ...               [ 1.,    0.  ],
    ...               [ 0.75,  0.25],
    ...               [ 0.55,  0.45],
    ...               [ 0.8,   0.2 ],
    ...               [ 0.95,  0.05]])
    >>> s = np.array([4, 5, 1, 7, 2, 6, 0, 8, 3, 9])
    >>> a, q = find_drainage_area_and_discharge_to_n(s, r, p)
    >>> a
    array([  1.    ,   2.575 ,   1.5   ,   1.    ,  10.    ,   5.2465,
             2.74  ,   2.845 ,   1.05  ,   1.    ])
    >>> q
    array([  1.    ,   2.575 ,   1.5   ,   1.    ,  10.    ,   5.2465,
             2.74  ,   2.845 ,   1.05  ,   1.    ])
    """
    # Number of points
    np = r.shape[0]
    q = r.shape[1]

    # Initialize the drainage_area and discharge arrays. Drainage area starts
    # out as the area of the cell in question, then (unless the cell has no
    # donors) grows from there. Discharge starts out as the cell's local runoff
    # rate times the cell's surface area.
    drainage_area = numpy.zeros(np) + node_cell_area
    discharge = numpy.zeros(np) + node_cell_area*runoff

    # Optionally zero out drainage area and discharge at boundary nodes
    if boundary_nodes is not None:
        drainage_area[boundary_nodes] = 0
        discharge[boundary_nodes] = 0

    # Iterate backward through the list, which means we work from upstream to
    # downstream.
    for i in range(np-1, -1, -1):
        donor = s[i]
        for v in range(q):
            recvr = r[donor, v]
            proportion = p[donor, v]
            if proportion > 0:
                if donor != recvr:
                    drainage_area[recvr] += proportion*drainage_area[donor]
                    discharge[recvr] += proportion*discharge[donor]

#        donors = s[i]
#        #print donors
#        recvrs = r[donors, :].flatten()
#
#        if (set(donors)-set(recvrs[recvrs!=-1]))==set(donors):
#            recvrs = r[donors, :].flatten()
#
#            unique_recvrs=numpy.unique(recvrs)
#
#            proportions = p[donors, :].flatten()
#
#            new_da=proportions*numpy.repeat(drainage_area[donors], q)
#            new_di=proportions*numpy.repeat(discharge[donors], q)
#
#            for u_r in unique_recvrs:
#                ur_ind=np.where(recvrs==u_r)
#
#                drainage_area[u_r] += numpy.sum(new_da[ur_ind])
#                discharge[u_r] += numpy.sum(new_di[ur_ind])

    return drainage_area, discharge


def flow_accumulation_to_n(receiver_nodes,
                           receiver_proportions,
                           node_cell_area=1.0,
                           runoff_rate=1.0,
                           boundary_nodes=None):

    """Calculate drainage area and (steady) discharge.

    Calculates and returns the drainage area and (steady) discharge at each
    node, along with a downstream-to-upstream ordered list (array) of node IDs.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.components.flow_accum.flow_accum_to_n import(
    ... flow_accumulation_to_n)
    >>> r = np.array([[ 1,  2],
    ...               [ 4,  5],
    ...               [ 1,  5],
    ...               [ 6,  2],
    ...               [ 4, -1],
    ...               [ 4, -1],
    ...               [ 5,  7],
    ...               [ 4,  5],
    ...               [ 6,  7],
    ...               [ 7,  8]])
    >>> p = np.array([[ 0.6,   0.4 ],
    ...               [ 0.85,  0.15],
    ...               [ 0.65,  0.35],
    ...               [ 0.9,   0.1 ],
    ...               [ 1.,    0.  ],
    ...               [ 1.,    0.  ],
    ...               [ 0.75,  0.25],
    ...               [ 0.55,  0.45],
    ...               [ 0.8,   0.2 ],
    ...               [ 0.95,  0.05]])
    >>> a, q, s = flow_accumulation_to_n(r, p)
    >>> a
    array([  1.    ,   2.575 ,   1.5   ,   1.    ,  10.    ,   5.2465,
             2.74  ,   2.845 ,   1.05  ,   1.    ])
    >>> q
    array([  1.    ,   2.575 ,   1.5   ,   1.    ,  10.    ,   5.2465,
             2.74  ,   2.845 ,   1.05  ,   1.    ])
    >>> s[0] == 4
    True
    >>> s[1] == 5
    True
    >>> s[9] == 9
    True
    >>> len(set([1, 7])-set(s[2:4]))
    0
    >>> len(set([2, 6])-set(s[4:6]))
    0
    >>> len(set([0, 3, 8])-set(s[6:9]))
    0
    """

    assert receiver_nodes.shape == receiver_proportions.shape, \
        'r and p arrays are not the same shape'

    s = make_ordered_node_array_to_n(receiver_nodes, receiver_proportions)
    # Note that this ordering of s DOES INCLUDE closed nodes. It really
    # shouldn't!
    # But as we don't have a copy of the grid accessible here, we'll solve this
    # problem as part of route_flow_dn.

    a, q = find_drainage_area_and_discharge_to_n(
        s, receiver_nodes, receiver_proportions, node_cell_area, runoff_rate,
        boundary_nodes)

    return a, q, s


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#! /usr/bin/env python
import os

import pytest

from landlab.core.messages import (format_message, error_message,
                                   warning_message, assert_or_print,
                                   split_paragraphs)


LOREM_IPSUM = os.linesep.join([
"Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.",
"",
"Pharetra pharetra massa massa ultricies mi quis hendrerit.",
"",
"Dictumst vestibulum rhoncus est pellentesque. Sed viverra tellus in hac habitasse platea dictumst vestibulum rhoncus.",
])


def test_split_paragraphs_cr():
    """Test splitting paragraphs with carriage returns."""
    text = """
Pharetra pharetra massa massa ultricies mi quis hendrerit.\r\rDictumst vestibulum rhoncus est pellentesque.
    """
    assert split_paragraphs(text, linesep='\r') == [
        "Pharetra pharetra massa massa ultricies mi quis hendrerit.",
        "Dictumst vestibulum rhoncus est pellentesque."]


def test_split_paragraphs_lf():
    """Test splitting paragraphs with line feeds."""
    text = """
Pharetra pharetra massa massa ultricies mi quis hendrerit.\n\nDictumst vestibulum rhoncus est pellentesque.
    """
    assert split_paragraphs(text, linesep='\n') == [
        "Pharetra pharetra massa massa ultricies mi quis hendrerit.",
        "Dictumst vestibulum rhoncus est pellentesque."]


def test_split_paragraphs_crlf():
    """Test splitting paragraphs with carriage returns and line feeds."""
    text = """
Pharetra pharetra massa massa ultricies mi quis hendrerit.\r\n\r\nDictumst vestibulum rhoncus est pellentesque.
    """
    assert split_paragraphs(text, linesep='\r\n') == [
        "Pharetra pharetra massa massa ultricies mi quis hendrerit.",
        "Dictumst vestibulum rhoncus est pellentesque."]


def test_empty_message():
    """Test formatting an empty string."""
    assert format_message('') == ''


def test_one_line():
    """Test a single line message."""
    assert format_message('lorem ipsum') == 'lorem ipsum'


def test_leading_whitespace():
    """Test a single line message."""
    assert format_message('   lorem ipsum') == 'lorem ipsum'


def test_one_long_line():
    """Test a line that needs to be wrapped."""
    msg = "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua."

    assert (
        format_message(msg) ==
        os.linesep.join([
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do",
            "eiusmod tempor incididunt ut labore et dolore magna aliqua."
        ])
    )


def test_multiline():
    msg = """lorem
ipsum
    """
    assert format_message(msg) == 'lorem ipsum'


def test_multiple_paragraphs():
    assert (
        format_message(LOREM_IPSUM) ==
        os.linesep.join([
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do",
            "eiusmod tempor incididunt ut labore et dolore magna aliqua.",
            "",
            "Pharetra pharetra massa massa ultricies mi quis hendrerit.",
            "",
            "Dictumst vestibulum rhoncus est pellentesque. Sed viverra tellus in",
            "hac habitasse platea dictumst vestibulum rhoncus.",
        ])
    )


def test_warning_message():
    msg = "Pharetra pharetra massa massa ultricies mi quis hendrerit."

    assert (
        warning_message(msg) ==
        os.linesep.join([
            "WARNING",
            "=======",
            "",
            "Pharetra pharetra massa massa ultricies mi quis hendrerit.",
        ])
    )


def test_error_message():
    msg = "Pharetra pharetra massa massa ultricies mi quis hendrerit."

    assert (
        error_message(msg) ==
        os.linesep.join([
            "ERROR",
            "=====",
            "",
            "Pharetra pharetra massa massa ultricies mi quis hendrerit.",
        ])
    )


def test_warning_message_is_none():
    assert (
        warning_message() ==
        os.linesep.join([
            "WARNING",
            "======="])
    )


def test_error_message_is_none():
    assert (
        error_message() ==
        os.linesep.join([
            "ERROR",
            "====="])
    )


def test_assert_or_pass():
    assert_or_print(True, onerror='pass')
    assert_or_print(False, onerror='pass')


def test_assert_or_warn():
    assert_or_print(True, onerror='warn')
    assert_or_print(False, onerror='warn')


def test_assert_or_raise():
    assert_or_print(True, onerror='raise')
    with pytest.raises(AssertionError):
        assert_or_print(False, onerror='raise')



import numpy as np
cimport numpy as np
cimport cython


@cython.boundscheck(False)
def deposit_or_erode(np.ndarray[np.float_t, ndim=2] layers, int n_layers,
                     np.ndarray[np.float_t, ndim=1] dz):
    cdef int n_stacks = layers.shape[1]
    cdef int top_ind = n_layers - 1
    cdef int col
    cdef int layer
    cdef double removed
    cdef double amount_to_remove

    for col in range(n_stacks):
        if dz[col] >= 0.:
            layers[top_ind, col] = dz[col]
        else:
            layers[top_ind, col] = 0.

            amount_to_remove = - dz[col]
            removed = 0.
            for layer in range(top_ind - 1, -1, -1):
                removed += layers[layer, col]
                layers[layer, col] = 0.
                if removed > amount_to_remove:
                    layers[layer, col] = removed - amount_to_remove
                    break



import numpy as np
cimport numpy as np
cimport cython

from libc.math cimport fabs, exp, cos, sin
# from libc.stdlib cimport abs


DTYPE = np.double
ctypedef np.double_t DTYPE_t


@cython.boundscheck(False)
def subside_load_1d(np.ndarray[DTYPE_t, ndim=1] x,
                    np.ndarray[DTYPE_t, ndim=2] loads,
                    DTYPE_t alpha,
                    DTYPE_t rigidity,
                    np.ndarray[DTYPE_t, ndim=2] out):
    cdef int n_rows = loads.shape[0]
    cdef int row

    for row in range(n_rows):
        subside_row(&x[0], len(x), &loads[row, 0], alpha, rigidity,
                    &out[row, 0])


@cython.cdivision(True)
cdef double * subside_row(double * x, int n_points, double * loads,
                          double alpha, double rigidity, double * out) nogil:
    cdef int col
    cdef double load

    for col in range(n_points):
        subside_point_load(x, n_points, x[col], loads[col], alpha, rigidity,
                           out)

    return out


@cython.cdivision(True)
cdef double * subside_point_load(double * x, int n_points, double x_at_load,
                                 double load, double alpha, double rigidity,
                                 double * out) nogil:
    cdef float c = load * alpha ** 3. / (8. * rigidity)
    cdef float dx
    cdef int col

    for col in range(n_points):
        dx = fabs(x[col] - x_at_load) / alpha
        out[col] += c * exp(- dx) * (cos(dx) + sin(dx))

    return out



#! /usr/env/python

"""
flow_director_to_one.py provides a private class to help create FlowDirectors.

Provides the _FlowDirectorToOne component which makes sure all model grid
fields are set up correctly.
"""

from landlab import FieldError
from landlab.components.flow_director.flow_director import _FlowDirector
import numpy
from landlab import BAD_INDEX_VALUE


class _FlowDirectorToOne(_FlowDirector):

    """
    Private class for creating components to calculate flow directions.

    This class is not meant to be used directly in modeling efforts. It
    inherits from the _FlowDirector class and builds on it to provide the
    functionality that all flow direction calculators need if they direct flow
    only to one nodes, as in steepest descent or D8 direction finding. It
    exists in contrast to the other intermediate flow director class
    _FlowDirectorToMany which provides equivalent functionality for flow
    direction algorithms such as D infinity or D trig that route flow from one
    cell to multiple nodes. As the primary difference between these two methods
    is the names of the fields they create and use, the primary function of
    this class is to create model grid fields.

    Specifically, it stores as ModelGrid fields:

    -  Node array of receivers (nodes that receive flow), or ITS OWN ID if
       there is no receiver: *'flow__receiver_node'*
    -  Node array of steepest downhill slopes:
       *'topographic__steepest_slope'*
    -  Node array containing ID of link that leads from each node to its
       receiver, or BAD_INDEX_VALUE if no link:
       *'flow__link_to_receiver_node'*
    -  Boolean node array of all local lows: *'flow__sink_flag'*

    The primary method of this class, :func:`run_one_step` is not implemented.

    Parameters
    ----------
    grid : ModelGrid
        A grid.
    surface : field name at node or array of length node
        The surface to direct flow across.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components.flow_director.flow_director_to_one import(
    ... _FlowDirectorToOne)
    >>> mg = RasterModelGrid((3,3), spacing=(1, 1))
    >>> mg.set_closed_boundaries_at_grid_edges(True, True, True, False)
    >>> _ = mg.add_field('topographic__elevation',
    ...                  mg.node_x + mg.node_y,
    ...                  at = 'node')
    >>> fd = _FlowDirectorToOne(mg, 'topographic__elevation')
    >>> fd.surface_values
    array([ 0.,  1.,  2.,  1.,  2.,  3.,  2.,  3.,  4.])
    >>> sorted(list(mg.at_node.keys()))
    ['flow__link_to_receiver_node', 'flow__receiver_node',
           'flow__sink_flag', 'topographic__elevation',
           'topographic__steepest_slope']
    """

    _name = 'FlowDirectorToOne'

    _input_var_names = ('topographic__elevation',
                        )

    _output_var_names = ('flow__receiver_node',
                         'topographic__steepest_slope',
                         'flow__link_to_receiver_node',
                         'flow__sink_flag',
                         )

    _var_units = {'topographic__elevation': 'm',
                  'flow__receiver_node': '-',
                  'topographic__steepest_slope': '-',
                  'flow__link_to_receiver_node': '-',
                  'flow__sink_flag': '-',
                  }

    _var_mapping = {'topographic__elevation': 'node',
                    'flow__receiver_node': 'node',
                    'topographic__steepest_slope': 'node',
                    'flow__link_to_receiver_node': 'node',
                    'flow__sink_flag': 'node',
                    }

    _var_doc = {
        'topographic__elevation': 'Land surface topographic elevation',
        'flow__receiver_node':
            'Node array of receivers (node that receives flow from current '
            'node)',
        'topographic__steepest_slope':
            'Node array of steepest *downhill* slopes',
        'flow__link_to_receiver_node':
            'ID of link downstream of each node, which carries the discharge',
        'flow__sink_flag': 'Boolean array, True at local lows',
    }

    def __init__(self, grid, surface):
        """Initialize the _FlowDirectorTo_One class."""
        # run init for the inherited class
        super(_FlowDirectorToOne, self).__init__(grid, surface)
        self.to_n_receivers = 'one'
        # initialize new fields
        try:
            self.receiver = grid.add_field('flow__receiver_node',
                                           BAD_INDEX_VALUE*grid.ones(at='node', dtype=int),
                                           at='node', dtype=int)
        except FieldError:
            self.receiver = grid.at_node['flow__receiver_node']

        try:
            self.steepest_slope = grid.add_zeros(
                'topographic__steepest_slope', at='node', dtype=float)
        except FieldError:
            self.steepest_slope = grid.at_node['topographic__steepest_slope']

        try:
            self.links_to_receiver = grid.add_field('flow__link_to_receiver_node',
                                                    BAD_INDEX_VALUE*grid.ones(at='node', dtype=int),
                                                    at='node', dtype=int)

        except FieldError:
            self.links_to_receiver = grid.at_node[
                'flow__link_to_receiver_node']

        grid.add_zeros('flow__sink_flag', at='node', dtype=numpy.int8,
                       noclobber=False)

    def run_one_step(self):
        """run_one_step is not implemented for this component."""
        raise NotImplementedError('run_one_step()')

    # set properties. These are the same for all DirectToOne Directors
    @property
    def node_receiving_flow(self):
        """Return the node id of the node receiving flow."""
        return self._grid['node']['flow__receiver_node']

    @property
    def node_steepest_slope(self):
        """Return the steepest link slope at a node."""
        return self._grid['node']['topographic__steepest_slope']

    @property
    def link_to_flow_receiving_node(self):
        """Return the link id along the link transporting flow."""
        return self._grid['node']['flow__link_to_receiver_node']

    @property
    def sink_flag(self):
        """Return the array with sink flags."""
        return self._grid['node']['flow__sink_flag']


if __name__ == '__main__':
    import doctest
    doctest.testmod()



import numpy as np
cimport numpy as np
cimport cython
from libc.stdlib cimport malloc, free


DTYPE = np.int
ctypedef np.int_t DTYPE_t


@cython.boundscheck(False)
def fill_perimeter_nodes(shape, np.ndarray[DTYPE_t, ndim=1] perimeter_nodes):
    cdef int n_rows = shape[0]
    cdef int n_cols = shape[1]
    cdef int n_nodes = n_rows * n_cols
    cdef int i
    cdef int node

    # Right edge
    i = 0
    for node in range(n_cols - 1, n_nodes - 1, n_cols):
        perimeter_nodes[i] = node
        i += 1

    # Top edge
    for node in range(n_nodes - 1, n_nodes - n_cols, - 1):
        perimeter_nodes[i] = node
        i += 1

    # Left edge
    for node in range((n_rows - 1) * n_cols, 0, - n_cols):
        perimeter_nodes[i] = node
        i += 1

    # Bottom edge
    for node in range(0, n_cols - 1):
        perimeter_nodes[i] = node
        i += 1


@cython.boundscheck(False)
def fill_hex_perimeter_nodes(shape,
                             np.ndarray[DTYPE_t, ndim=1] perimeter_nodes):
    cdef int n_rows = shape[0]
    cdef int n_cols = shape[1]
    cdef int n_bottom_rows = (n_rows + (n_rows + 1) % 2) // 2 + 1
    cdef int i, i0
    cdef int node
    cdef int row
    cdef int * nodes_per_row = <int *>malloc(n_rows * sizeof(int))

    try:
        nodes_per_row[0] = n_cols
        for row in range(1, n_bottom_rows):
            nodes_per_row[row] = nodes_per_row[row - 1] + 1
        for row in range(n_bottom_rows, n_rows):
            nodes_per_row[row] = nodes_per_row[row - 1] - 1

        # Right edge
        perimeter_nodes[0] = n_cols - 1
        row = 1
        for i in range(1, n_rows):
            perimeter_nodes[i] = perimeter_nodes[i - 1] + nodes_per_row[row]
            row += 1

        # Top edge
        i0 = i + 1
        for i in range(i0, i0 + nodes_per_row[n_rows - 1] - 1):
            perimeter_nodes[i] = perimeter_nodes[i - 1] - 1

        # Left edge
        i0 = i + 1
        row = n_rows - 2
        for i in range(i0, i0 + n_rows - 1):
            perimeter_nodes[i] = perimeter_nodes[i - 1] - nodes_per_row[row]
            row -= 1

        # Bottom edge
        i0 = i + 1
        for i in range(i0, i0 + n_cols - 2):
            perimeter_nodes[i] = perimeter_nodes[i - 1] + 1
    finally:
        free(nodes_per_row)


@cython.boundscheck(False)
def fill_hex_xy_of_node(shape,
                        np.ndarray[np.double_t, ndim=1] x_of_node,
                        np.ndarray[np.double_t, ndim=1] y_of_node):
    """Get x and y coordinates for each node."""
    cdef int n_nodes = x_of_node.size
    cdef int n_cols = shape[1]
    cdef int node
    cdef int row
    cdef int col
    cdef int offset
    cdef int n
    cdef double x0

    node = 0
    x0 = 0.
    for row in range(shape[0] / 2 + 1):
        for col in range(n_cols):
            x_of_node[node] = x0 + col
            y_of_node[node] = row
            node += 1
        x0 -= .5
        n_cols += 1

    x0 += 1.
    n_cols -= 2
    for row in range(shape[0] / 2 + 1, shape[0]):
        for col in range(n_cols):
            x_of_node[node] = x0 + col
            y_of_node[node] = row
            node += 1
        x0 += .5
        n_cols -= 1


@cython.boundscheck(False)
def fill_xy_of_node(shape,
                    np.ndarray[np.double_t, ndim=1] x_of_node,
                    np.ndarray[np.double_t, ndim=1] y_of_node):
    """Get x and y coordinates for each node."""
    cdef int n_nodes = x_of_node.size
    cdef int stride = 2 * shape[1]
    cdef int n_cols = shape[1]
    cdef int row
    cdef int offset
    cdef int n

    for offset in range(0, n_nodes, stride):
        row = (offset // stride) * 2
        for n in range(offset, offset + n_cols):
            x_of_node[n] = n - offset
            y_of_node[n] = row

    for offset in range(shape[1], n_nodes, stride):
        row = (offset // stride) * 2 + 1
        for n in range(offset, offset + n_cols):
            x_of_node[n] = n - offset + .5
            y_of_node[n] = row


@cython.boundscheck(False)
def get_xy_of_node(shape,
                   np.ndarray[np.double_t, ndim=1] x_of_node,
                   np.ndarray[np.double_t, ndim=1] y_of_node):
    """Get x and y coordinates for each node."""
    cdef int n_nodes = x_of_node.size
    cdef int stride = 2 * shape[1] + 1
    cdef int n_cols = shape[1]
    cdef int row
    cdef int offset
    cdef int n

    for offset in range(0, n_nodes, stride):
        row = (offset // stride) * 2
        for n in range(offset, offset + n_cols):
            x_of_node[n] = n - offset + .5
            y_of_node[n] = row

    for offset in range(shape[1], n_nodes, stride):
        row = (offset // stride) * 2 + 1
        for n in range(offset, offset + n_cols + 1):
            x_of_node[n] = n - offset
            y_of_node[n] = row


@cython.boundscheck(False)
def get_nodes_at_link(shape,
                   np.ndarray[DTYPE_t, ndim=2] nodes_at_link):
    """Get nodes at the tail and head of each node."""
    cdef int n_links = nodes_at_link.shape[0]
    cdef int n_short_rows = (shape[0] + 1) // 2
    cdef int n_long_rows = shape[0] // 2
    cdef int n_cols = shape[1]
    cdef int n_rows = shape[0]
    cdef int stride = 2 * n_cols + 1
    cdef int row
    cdef int n
    cdef int link

    link = 0
    for row in range(n_rows):
        node = row * stride // 2
        if row % 2 == 0:
            for n in range(n_cols - 1):
                nodes_at_link[link, 0] = n + node
                nodes_at_link[link, 1] = n + node + 1
                link += 1
        else:
            for n in range(n_cols):
                nodes_at_link[link, 0] = n + node
                nodes_at_link[link, 1] = n + node + 1
                link += 1

        if row == n_rows - 1:
            break

        if row % 2 == 0:
            for n in range(n_cols):
                nodes_at_link[link, 0] = n + node
                nodes_at_link[link, 1] = n + node + n_cols
                link += 1
                nodes_at_link[link, 0] = n + node
                nodes_at_link[link, 1] = n + node + n_cols + 1
                link += 1
        else:
            nodes_at_link[link, 0] = node
            nodes_at_link[link, 1] = node + n_cols + 1
            link += 1
            for n in range(1, n_cols):
                nodes_at_link[link, 0] = n + node
                nodes_at_link[link, 1] = n + node + n_cols
                link += 1
                nodes_at_link[link, 0] = n + node
                nodes_at_link[link, 1] = n + node + n_cols + 1
                link += 1
            nodes_at_link[link, 0] = node + n_cols
            nodes_at_link[link, 1] = node + n_cols + n_cols
            link += 1


@cython.boundscheck(False)
def get_links_at_patch(shape,
                       np.ndarray[DTYPE_t, ndim=2] links_at_patch):
    """Get links that bound each patch."""
    cdef int n_patches = links_at_patch.shape[0]
    cdef int n_short_rows = (shape[0] + 1) // 2
    cdef int n_long_rows = shape[0] // 2
    cdef int n_cols = shape[1]
    cdef int n_rows = shape[0]
    cdef int row
    cdef int n
    cdef int patch
    cdef int link
    cdef int link_at_row

    patch = 0
    link_at_row = n_cols - 1
    for row in range(n_rows - 1):
        if row % 2 == 0:
            link = link_at_row + 2
            n_patches = n_cols - 1
        else:
            link = link_at_row + 1
            n_patches = n_cols

        # upward pointing
        links_at_patch[patch, 0] = link
        links_at_patch[patch, 1] = link - 1
        links_at_patch[patch, 2] = link - (n_cols + 1)
        patch += 1
        for n in range(n_patches - 1):
            links_at_patch[patch, 0] = links_at_patch[patch - 1, 0] + 2
            links_at_patch[patch, 1] = links_at_patch[patch - 1, 1] + 2
            links_at_patch[patch, 2] = links_at_patch[patch - 1, 2] + 1
            patch += 1

        if row % 2 == 0:
            link = link_at_row
            n_patches = n_cols
        else:
            link = link_at_row + 1
            n_patches = n_cols - 1

        # downward pointing
        if row % 2 == 0:
            links_at_patch[patch, 0] = link + 2 * n_cols
        else:
            links_at_patch[patch, 0] = link + 2 * n_cols - 1
        links_at_patch[patch, 1] = link
        links_at_patch[patch, 2] = link + 1
        patch += 1
        for n in range(n_patches - 1):
            links_at_patch[patch, 0] = links_at_patch[patch - 1, 0] + 1
            links_at_patch[patch, 1] = links_at_patch[patch - 1, 1] + 2
            links_at_patch[patch, 2] = links_at_patch[patch - 1, 2] + 2
            patch += 1

        if row % 2 == 0:
            link_at_row += 3 * n_cols
        else:
            link_at_row += 3 * n_cols - 1



from .generate_overland_flow_Bates import OverlandFlowBates
from .generate_overland_flow_deAlmeida import OverlandFlow
from .generate_overland_flow_implicit_kinwave import KinwaveImplicitOverlandFlow
from .generate_overland_flow_kinwave import KinwaveOverlandFlowModel

__all__ = ['OverlandFlowBates',
           'OverlandFlow',
	   'KinwaveImplicitOverlandFlow',
       'KinwaveOverlandFlowModel']



#! /usr/env/python
"""Simple hexagonal Landlab cellular automaton

This file defines the OrientedHexCTS class, which is a sub-class of
CellLabCTSModel that implements a simple, non-oriented, hex-grid
CA. Like its parent class, OrientedHexCTS implements a continuous-time,
stochastic, pair-based CA. The hex grid has 3 principal directions, rather
than 2 for a raster. Hex grids are often used in CA models because of their
symmetry.

Created GT Sep 2014
"""
import numpy as np

from .celllab_cts import CellLabCTSModel
from ..grid import HexModelGrid


class OrientedHexCTS(CellLabCTSModel):

    """Oriented hex-grid CellLab-CTS model.

    OrientedHexCTS constructor: sets number of orientations to 3 and calls
    base-class constructor.

    Parameters
    ----------
    model_grid : Landlab ModelGrid object
        Reference to the model's grid
    node_state_dict : dict
        Keys are node-state codes, values are the names associated with
        these codes
    transition_list : list of Transition objects
        List of all possible transitions in the model
    initial_node_states : array of ints (x number of nodes in grid)
        Starting values for node-state grid
    prop_data : array (x number of nodes in grid), optional
        Array of properties associated with each node/cell
    prop_reset_value : number or object, optional
        Default or initial value for a node/cell property (e.g., 0.0).
        Must be same type as *prop_data*.

    Examples
    --------
    >>> from landlab import HexModelGrid
    >>> from landlab.ca.oriented_hex_cts import OrientedHexCTS
    >>> from landlab.ca.celllab_cts import Transition

    >>> mg = HexModelGrid(4, 3, 1.0)
    >>> nsd = {0 : 'yes', 1 : 'no'}
    >>> xnlist = []
    >>> xnlist.append(Transition((0,1,0), (1,1,0), 1.0, 'frogging'))
    >>> nsg = mg.add_zeros('node', 'node_state_grid')
    >>> ohcts = OrientedHexCTS(mg, nsd, xnlist, nsg)
    """

    def __init__(self, model_grid, node_state_dict, transition_list,
                 initial_node_states, prop_data=None, prop_reset_value=None):
        """Initialize a OrientedHexCTS.

        OrientedHexCTS constructor: sets number of orientations to 3 and calls
        base-class constructor.

        Parameters
        ----------
        model_grid : Landlab ModelGrid object
            Reference to the model's grid
        node_state_dict : dict
            Keys are node-state codes, values are the names associated with
            these codes
        transition_list : list of Transition objects
            List of all possible transitions in the model
        initial_node_states : array of ints (x number of nodes in grid)
            Starting values for node-state grid
        prop_data : array (x number of nodes in grid), optional
            Array of properties associated with each node/cell
        prop_reset_value : number or object, optional
            Default or initial value for a node/cell property (e.g., 0.0).
            Must be same type as *prop_data*.
        """

        # Make sure caller has sent the right grid type
        if not isinstance(model_grid, HexModelGrid):
            raise TypeError('model_grid must be a Landlab HexModelGrid')

        # Somehow test to make sure the grid links have been re-oriented to
        # point up/right (-45 to +135 degrees clockwise relative to vertical).
        # Such orientation is ensured when the argument reorient_grid=True is
        # passed to the hex grid constructor.

        # Define the number of distinct cell-pair orientations: here 3,
        # representing
        self.number_of_orientations = 3

        # Call the LandlabCellularAutomaton.__init__() method to do the rest of
        # the initialization
        super(OrientedHexCTS, self).__init__(model_grid, node_state_dict,
                                             transition_list,
                                             initial_node_states, prop_data,
                                             prop_reset_value)

    def setup_array_of_orientation_codes(self):
        """
        Creates and configures an array that contain the orientation code for
        each active link (and corresponding cell pair).

        Notes
        -----
        **Creates**:

        * ``self.active_link_orientation``: 1D numpy array

        This overrides the method of the same name in celllab_cts.py. If the
        hex grid is oriented such that one of the 3 axes is vertical (a
        'vertical' grid), then the three orientations are:

        * 0 = vertical (0 degrees clockwise from vertical)
        * 1 = right and up (60 degrees clockwise from vertical)
        * 2 = right and down (120 degrees clockwise from vertical)

        If the grid is oriented with one principal axis horizontal
        ('horizontal' grid), then the orientations are:

        * 0 = up and left (30 degrees counter-clockwise from vertical)
        * 1 = up and right (30 degrees clockwise from vertical)
        * 2 = horizontal (90 degrees clockwise from vertical)
        """
        self.link_orientation = np.zeros(self.grid.number_of_links,
                                         dtype=np.int8)
        for i in range(self.grid.number_of_links):
            dy = self.grid.node_y[self.grid.node_at_link_head[
                i]] - self.grid.node_y[self.grid.node_at_link_tail[i]]
            dx = self.grid.node_x[self.grid.node_at_link_head[
                i]] - self.grid.node_x[self.grid.node_at_link_tail[i]]
            if dx <= 0.:
                self.link_orientation[i] = 0
            elif dy <= 0.:
                self.link_orientation[i] = 2
            elif dx > 0. and dy > 0.:
                self.link_orientation[i] = 1
            else:
                assert (False), 'Non-handled link orientation case'


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
stream_power_smooth_threshold.py: Defines the StreamPowerSmoothThresholdEroder,
which is a version of the FastscapeEroder (derived from it).

StreamPowerSmoothThresholdEroder uses a mathematically smooth threshold
formulation, rather than one with a singularity. The erosion rate is defined as
follows:

$\omega = K A^m S$

$E = \omega - \omega_c \left[ 1 - \exp ( -\omega / \omega_c ) \right]$

Created on Sat Nov 26 08:36:49 2016

@author: gtucker
"""

if __name__ == '__main__':
    from landlab.components import FastscapeEroder
else:
    from .fastscape_stream_power import FastscapeEroder
import numpy as np
from .cfuncs import smooth_stream_power_eroder_solver
from landlab import BAD_INDEX_VALUE
UNDEFINED_INDEX = BAD_INDEX_VALUE

class StreamPowerSmoothThresholdEroder(FastscapeEroder):
    """Stream erosion component with smooth threshold function.

    Parameters
    ----------
    grid : ModelGrid
        A grid.
    K_sp : float, array, or field name
        K in the stream power equation (units vary with other parameters).
    m_sp : float, optional
        m in the stream power equation (power on drainage area).
    n_sp : float, optional, ~ 0.5<n_sp<4.
        n in the stream power equation (power on slope). NOTE: NOT PRESENTLY
        HONORED BY StreamPowerSmoothThresholdEroder (TODO)
    threshold_sp : float (TODO: array, or field name)
        The threshold stream power.
    rainfall_intensity : float; optional
        NOT PRESENTLY HONORED (TODO)

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> rg = RasterModelGrid((3, 4), 1.0)
    >>> rg.set_closed_boundaries_at_grid_edges(False, True, True, True)
    >>> z = rg.add_zeros('node', 'topographic__elevation')
    >>> z[5] = 2.0
    >>> z[6] = 1.0
    >>> from landlab.components import FlowRouter
    >>> fr = FlowRouter(rg, method='D4')
    >>> fr.run_one_step()
    >>> from landlab.components import StreamPowerSmoothThresholdEroder
    >>> sp = StreamPowerSmoothThresholdEroder(rg, K_sp=1.0)
    >>> sp.thresholds
    1.0
    >>> sp.run_one_step(dt=1.0)
    >>> import numpy as np
    >>> np.round(z[5:7], 3)
    array([ 1.646,  0.667])
    >>> z[5] = 2.0
    >>> z[6] = 1.0
    >>> import numpy as np
    >>> q = np.zeros(rg.number_of_nodes) + 0.25
    >>> q[6] = 100.0
    >>> sp = StreamPowerSmoothThresholdEroder(rg, K_sp=1.0, use_Q=q)
    >>> sp.run_one_step(dt=1.0)
    >>> np.round(z[5:7], 3)
    array([ 1.754,  0.164])
    """

    def __init__(self, grid, K_sp=None, m_sp=0.5, n_sp=1., threshold_sp=1.,
                 rainfall_intensity=1., use_Q=None, **kwargs):
        """Initialize StreamPowerSmoothThresholdEroder."""
        if n_sp != 1.0:
            raise ValueError(('StreamPowerSmoothThresholdEroder currently only '
                              'supports n_sp = 1'))

        # Call base-class init
        super(StreamPowerSmoothThresholdEroder,
              self).__init__(grid, K_sp=K_sp, m_sp=m_sp, n_sp=n_sp,
                             threshold_sp=threshold_sp,
                             rainfall_intensity=rainfall_intensity, **kwargs)

        # Handle "use_Q" option (ideally should be done by base class, but
        # FastscapeEroder, unlike StreamPowerEroder, lacks this option)
        if use_Q is not None:
            if type(use_Q) is str:  # if str, assume it's a field name
                self.area_or_discharge = self._grid.at_node[use_Q]
            elif type(use_Q) is np.ndarray:  # if array, use it
                self.area_or_discharge = use_Q
            else:
                print('Warning: use_Q must be field name or array')
                self.area_or_discharge = self._grid.at_node['drainage_area']
        else:
            self.area_or_discharge = self._grid.at_node['drainage_area']

        # Arrays with parameters for use in implicit solver
        self.gamma = grid.empty(at='node')
        self.delta = grid.empty(at='node')


    def run_one_step(self, dt, flooded_nodes=None, runoff_rate=None, **kwds):
        """Run one forward iteration of duration dt.

        Parameters
        ----------
        dt : float
            Time step size
        flooded_nodes : ndarray of int (optional)
            Indices of nodes in lakes/depressions
        runoff_rate : (not used yet)
            (to be added later)

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> rg = RasterModelGrid((3, 3), 1.0)
        >>> rg.set_closed_boundaries_at_grid_edges(False, True, True, True)
        >>> z = rg.add_zeros('node', 'topographic__elevation')
        >>> z[4] = 1.0
        >>> from landlab.components import FlowRouter
        >>> fr = FlowRouter(rg, method='D4')
        >>> fr.run_one_step()
        >>> from landlab.components import StreamPowerSmoothThresholdEroder
        >>> sp = StreamPowerSmoothThresholdEroder(rg, K_sp=1.0)
        >>> sp.run_one_step(dt=1.0)
        >>> sp.alpha
        array([ 0.,  0.,  0.,  0.,  1.,  0.,  0.,  0.,  0.])
        >>> sp.gamma
        array([ 0.,  0.,  0.,  0.,  1.,  0.,  0.,  0.,  0.])
        >>> sp.delta
        array([ 0.,  0.,  0.,  0.,  1.,  0.,  0.,  0.,  0.])
        """

        # Set up needed arrays
        #
        # Get shorthand for elevation field ("z"), and for up-to-downstream
        # ordered array of node IDs ("upstream_order_IDs")
        node_id = np.arange(self._grid.number_of_nodes)
        upstream_order_IDs = self._grid['node']['flow__upstream_node_order']
        z = self._grid['node']['topographic__elevation']
        flow_receivers = self._grid['node']['flow__receiver_node']

        # Get an array of flow-link length for each node that has a defined
        # receiver (i.e., that drains to another node).
        defined_flow_receivers = np.not_equal(self._grid['node'][
            'flow__link_to_receiver_node'], UNDEFINED_INDEX)
        defined_flow_receivers[flow_receivers == node_id] = False
        if flooded_nodes is not None:
            defined_flow_receivers[flooded_nodes] = False
        flow_link_lengths = self._grid.length_of_d8[
            self._grid.at_node['flow__link_to_receiver_node'][defined_flow_receivers]]

        # (Later on, add functionality for a runoff rate, or discharge, or
        # variable K)

        # Handle possibility of spatially varying K
        if type(self.K) is np.ndarray:
            K = self.K[defined_flow_receivers]
        else:
            K = self.K

        # Handle possibility of spatially varying threshold
        if isinstance(self.thresholds, np.ndarray):
            thresh = self.thresholds[defined_flow_receivers]
        else:
            thresh = self.thresholds

        # Set up alpha, beta, delta arrays
        #
        #   First, compute drainage area or discharge raised to the power m.
        np.power(self.area_or_discharge, self.m,
                 out=self.A_to_the_m)

        #   Alpha
        self.alpha[defined_flow_receivers==False] = 0.0
        self.alpha[defined_flow_receivers] = (K * dt
            * self.A_to_the_m[defined_flow_receivers] / flow_link_lengths)

        #   Gamma
        self.gamma[defined_flow_receivers==False] = 0.0
        self.gamma[defined_flow_receivers] = dt * thresh

        #   Delta
        self.delta[defined_flow_receivers==False] = 0.0
        if isinstance(self.thresholds, np.ndarray):
            self.delta[defined_flow_receivers] = ((K
                * self.A_to_the_m[defined_flow_receivers] )
                / (thresh * flow_link_lengths))

            self.delta[defined_flow_receivers][thresh == 0.0] = 0.0
        else:
            if thresh == 0:
                self.delta[defined_flow_receivers] = 0.0
            else:
                self.delta[defined_flow_receivers] = ((K
                    * self.A_to_the_m[defined_flow_receivers] )
                    / (thresh * flow_link_lengths))

        # Iterate over nodes from downstream to upstream, using scipy's
        # 'newton' function to find new elevation at each node in turn.
        smooth_stream_power_eroder_solver(upstream_order_IDs, flow_receivers,
                                         z, self.alpha, self.gamma, self.delta)

#if __name__ == '__main__':
#    from landlab import RasterModelGrid
#    rg = RasterModelGrid((3, 4), 1.0)
#    z = rg.add_zeros('node', 'topographic__elevation')
#    rg.set_closed_boundaries_at_grid_edges(False, True, True, True)
#    z[5] = 2.0
#    z[6] = 1.0
#    from landlab.components import FlowRouter
#    fr = FlowRouter(rg, method='D4')
#    fr.run_one_step()
#    sp = StreamPowerSmoothThresholdEroder(rg, K_sp=1.0)
#    sp.run_one_step(dt=1.0)
#    print(z)



import numpy as np
import xarray as xr
from scipy.spatial import Voronoi
from scipy.spatial import Delaunay

from ..graph import Graph
from ...core.utils import as_id_array


from ..ugrid import (MESH_ATTRS, update_node_coords, update_nodes_at_link,
                     update_links_at_patch)


def remove_bad_patches(max_node_spacing, nodes_at_patch, neighbors_at_patch):
    from .ext.delaunay import remove_tris

    max_node_dist = np.ptp(nodes_at_patch, axis=1)
    bad_patches = as_id_array(np.where(max_node_dist > max_node_spacing)[0])

    if len(bad_patches) > 0:
        remove_tris(nodes_at_patch, neighbors_at_patch, bad_patches)
        nodes_at_patch = nodes_at_patch[:-len(bad_patches), :]
        neighbors_at_patch = neighbors_at_patch[:-len(bad_patches), :]

    return nodes_at_patch, neighbors_at_patch


def setup_links_and_patches(node_y_and_x, max_node_spacing=None):
    from .ext.delaunay import _setup_links_at_patch, remove_tris

    delaunay = Delaunay(list(zip(node_y_and_x[1], node_y_and_x[0])))

    nodes_at_patch = np.asarray(delaunay.simplices, dtype=int)
    neighbors_at_patch = np.asarray(delaunay.neighbors, dtype=int)

    if max_node_spacing is not None:
        nodes_at_patch, neighbors_at_patch = remove_bad_patches(
            max_node_spacing, nodes_at_patch, neighbors_at_patch)

    n_patches = len(nodes_at_patch)
    n_shared_links = np.count_nonzero(neighbors_at_patch > -1)
    n_links = 3 * n_patches - n_shared_links // 2

    links_at_patch = np.empty((n_patches, 3), dtype=int)
    nodes_at_link = np.empty((n_links, 2), dtype=int)

    _setup_links_at_patch(nodes_at_patch,
                          neighbors_at_patch,
                          nodes_at_link, links_at_patch)

    return nodes_at_link, links_at_patch


def ugrid_from_voronoi(node_y_and_x, max_node_spacing=None):
    from .ext.delaunay import _setup_links_at_patch, remove_tris

    ugrid = xr.Dataset({'mesh': xr.DataArray(data=1, attrs=MESH_ATTRS)})

    nodes_at_link, links_at_patch = (
        setup_links_and_patches(node_y_and_x,
                                max_node_spacing=max_node_spacing))

    update_node_coords(ugrid, node_y_and_x)
    update_nodes_at_link(ugrid, nodes_at_link)
    update_links_at_patch(ugrid, links_at_patch)

    return ugrid


class VoronoiGraph(Graph):

    """Graph of a voronoi grid.

    Examples
    --------
    >>> from landlab.graph import VoronoiGraph
    """

    def __init__(self, node_y_and_x, **kwds):
        """Create a voronoi grid.

        Parameters
        ----------
        nodes : tuple of array_like
            Coordinates of every node. First *y*, then *x*.

        Examples
        --------
        >>> from landlab.graph import VoronoiGraph
        >>> node_x = [0, 1, 2,
        ...           1, 2, 3]
        >>> node_y = [0, 0, 0,
        ...           2, 2, 2]
        >>> graph = VoronoiGraph((node_y, node_x))
        >>> graph.x_of_node
        array([ 0.,  1.,  2.,  1.,  2.,  3.])
        >>> graph.y_of_node
        array([ 0.,  0.,  0.,  2.,  2.,  2.])
        >>> graph.nodes_at_link # doctest: +NORMALIZE_WHITESPACE
        array([[0, 1], [1, 2],
               [0, 3], [1, 3], [1, 4], [2, 4], [2, 5],
               [3, 4], [4, 5]])
        >>> graph.links_at_node # doctest: +NORMALIZE_WHITESPACE
        array([[ 0,  2, -1, -1], [ 1,  4,  3,  0], [ 6,  5,  1, -1],
               [ 7,  2,  3, -1], [ 8,  7,  4,  5], [ 8,  6, -1, -1]])
        >>> graph.links_at_patch # doctest: +NORMALIZE_WHITESPACE
        array([[3, 2, 0], [5, 4, 1], [4, 7, 3], [6, 8, 5]])
        >>> graph.nodes_at_patch # doctest: +NORMALIZE_WHITESPACE
        array([[3, 0, 1], [4, 1, 2], [4, 3, 1], [5, 4, 2]])
        """
        max_node_spacing = kwds.pop('max_node_spacing', None)
        mesh = ugrid_from_voronoi(node_y_and_x,
                                  max_node_spacing=max_node_spacing)
        Graph.__init__(self, mesh, **kwds)



import warnings

import numpy as np
from ...utils.jaggedarray import unravel
from ...core.utils import as_id_array


# def get_links_at_node(graph):
#     (links_at_node,
#      _,
#      offset_to_node) = calc_links_at_node(graph.nodes_at_link)
# 
#     return unravel(links_at_node, offset_to_node)


def get_link_dirs_at_node(graph):
    (_,
     link_dirs_at_node,
     offset_to_node) = calc_links_at_node(graph.nodes_at_link)

    return unravel(link_dirs_at_node, offset_to_node)


def get_links_at_node(graph, sort=False):
    """Set up data structures for node-to-link connectivity.

    Parameters
    ----------
    graph : Graph
        A `Graph`.

    nodes_at_link: ndarray of int
        Nodes at either end of a link (tail node, then head node).
    number_of_nodes: int, optional
        The total number of nodes. If not given, use the largest node in
        *nodes_at_link*.

    Returns
    -------
    tuple of ndarray
        Tuple of *links_at_node* and *link_dirs_at_node*.
    """
    # from .cfuncs import _setup_links_at_node
    from .ext.at_node import get_links_at_node

    sorted_by_node = np.argsort(graph.nodes_at_link.flat)

    node_count = np.bincount(graph.nodes_at_link.flat)
    number_of_nodes = graph.number_of_nodes

    max_node_count = np.max(node_count)

    link_dirs_at_node = np.full((number_of_nodes, max_node_count), 0,
                                dtype=int)
    links_at_node = np.full((number_of_nodes, max_node_count), -1, dtype=int)

    get_links_at_node(graph.nodes_at_link, links_at_node, link_dirs_at_node)

    if sort:
        sort_links_at_node_by_angle(links_at_node, link_dirs_at_node,
                                    graph.angle_of_link, inplace=True)

    return links_at_node, link_dirs_at_node


def sort_links_at_node_by_angle(links_at_node, link_dirs_at_node,
                                angle_of_link, inplace=True):
    """Sort links as spokes about a hub.

    Parameters
    ----------
    links_at_node : ndarray of int, shape `(n_nodes, max_links_per_node)`
        Links entering or leaving each node.
    link_dirs_at_node : ndarray of int, shape `(n_nodes, max_links_per_node)`
        Direction of links entering or leaving each node.
    angle_of_link : ndarray of float, shape `(n_links, )`
        Angle (in radians) of each link as measured from its head to tail.

    Returns
    -------
    tuple of (links_at_node, link_dirs_at_node)
        The sorted arrays. If `inplace` is `True`, these are the input
        arrays.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.graph.object.at_node import sort_links_at_node_by_angle

    ::
        (1) - 1 -> (3)
         |          ^
         2          3
         V          |
        (0) - 0 -> (2)

    >>> links_at_node = [[2, 0], [1, 2], [3, 0], [3, 1]]
    >>> link_dirs_at_node = [[1, -1], [-1, -1], [-1, 1], [1, 1]]
    >>> angle_of_link = np.array([0., 0., -90., 90.]) * np.pi / 180.

    >>> out = sort_links_at_node_by_angle(links_at_node, link_dirs_at_node,
    ...     angle_of_link)

    The first item of the returned tuple is links at each node sorted
    counterclockwise by angle.

    >>> out[0] # doctest: +NORMALIZE_WHITESPACE
    array([[0, 2], [2, 1], [3, 0],  [1, 3]])

    The second item is the direction of the link (entering or leaving).

    >>> out[1] # doctest: +NORMALIZE_WHITESPACE
    array([[-1,  1], [-1, -1], [-1,  1], [ 1,  1]])

    Because the input arrays are lists, not numpy arrays, the sort is not
    in-place.

    >>> out[0] is not links_at_node
    True

    >>> links_at_node = np.asarray([[2, 0], [1, 2], [3, 0], [3, 1]])
    >>> link_dirs_at_node = np.asarray([[1, -1], [-1, -1], [-1, 1], [1, 1]])
    >>> angle_of_link = np.array([0., 0., -90., 90.]) * np.pi / 180.

    >>> _ = sort_links_at_node_by_angle(links_at_node, link_dirs_at_node,
    ...     angle_of_link)
    >>> links_at_node # doctest: +NORMALIZE_WHITESPACE
    array([[0, 2], [2, 1], [3, 0],  [1, 3]])
    >>> link_dirs_at_node # doctest: +NORMALIZE_WHITESPACE
    array([[-1,  1], [-1, -1], [-1,  1], [ 1,  1]])
    """
    from .ext.at_node import reorder_links_at_node

    out = (np.asarray(links_at_node, dtype=int),
           np.asarray(link_dirs_at_node, dtype=int))

    if inplace:
        if out[0] is not links_at_node or out[1] is not link_dirs_at_node:
            inplace = False
            warnings.warn('input must be ndarray of int for in-place sort')

    if not inplace:
        if out[0] is links_at_node:
            out[0] = links_at_node.copy()
        if out[1] is link_dirs_at_node:
            out[1] = link_dirs_at_node.copy()

    links_at_node, link_dirs_at_node = out

    outward_angle = angle_of_link[links_at_node]

    links_entering = np.where(link_dirs_at_node == 1)
    outward_angle[links_entering] += np.pi
    outward_angle[outward_angle >= 2 * np.pi] -= 2 * np.pi

    outward_angle[np.where(link_dirs_at_node == 0)] = 4 * np.pi

    sorted_links = as_id_array(np.argsort(outward_angle))

    reorder_links_at_node(links_at_node, sorted_links)
    reorder_links_at_node(link_dirs_at_node, sorted_links)

    return links_at_node, link_dirs_at_node



# -*- coding: utf-8 -*-
"""
Unit tests for source_tracking_algorithm.py
@author: Sai Nudurupati & Erkan Istanbulluoglu
"""

import numpy as np

from landlab import RasterModelGrid
from landlab.components import FlowRouter
from landlab.utils import (track_source,
                           find_unique_upstream_hsd_ids_and_fractions)

def test_track_source():
    """Unit tests for track_source().
    """
    grid = RasterModelGrid((5, 5), spacing=(1., 1.))
    grid.at_node['topographic__elevation'] = np.array([5., 5., 5., 5., 5.,
                                                       5., 4., 5., 1., 5.,
                                                       0., 3., 5., 3., 0.,
                                                       5., 4., 5., 2., 5.,
                                                       5., 5., 5., 5., 5.])
    grid.status_at_node[10] = 0
    grid.status_at_node[14] = 0
    fr = FlowRouter(grid)
    fr.route_flow()
    r = grid.at_node['flow__receiver_node']
    assert r[6] == 10
    assert r[7] == 8
    assert r[18] == 14
    hsd_ids = np.empty(grid.number_of_nodes, dtype=int)
    hsd_ids[:] = 1
    hsd_ids[2:5] = 0
    hsd_ids[7:10] = 0
    (hsd_upstr, flow_accum) = track_source(grid, hsd_ids)
    assert hsd_upstr[8] == [1, 0, 0]
    assert hsd_upstr[14] == [1, 1, 1, 1, 0, 0, 1]
    assert flow_accum[14] == 7

def test_find_unique_upstream_hsd_ids_and_fractions():
    """Unit tests find_unique_upstream_hsd_ids_and_fractions().
    """
    grid = RasterModelGrid((5, 5), spacing=(1., 1.))
    grid.at_node['topographic__elevation'] = np.array([5., 5., 5., 5., 5.,
                                                       5., 4., 5., 1., 5.,
                                                       0., 3., 5., 3., 0.,
                                                       5., 4., 5., 2., 5.,
                                                       5., 5., 5., 5., 5.])
    grid.status_at_node[10] = 0
    grid.status_at_node[14] = 0
    fr = FlowRouter(grid)
    fr.route_flow()
    hsd_ids = np.empty(grid.number_of_nodes, dtype=int)
    hsd_ids[:] = 1
    hsd_ids[2:5] = 0
    hsd_ids[7:10] = 0
    (hsd_upstr, flow_accum) = track_source(grid, hsd_ids)
    (uniq_ids, coeff) = find_unique_upstream_hsd_ids_and_fractions(hsd_upstr)
    np.testing.assert_almost_equal(
        np.sort(np.array(coeff[8])), np.array([0.33333333, 0.66666667]))



import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal
from landlab import BAD_INDEX_VALUE as XX

from landlab import RasterModelGrid
import landlab.grid.mappers as maps


class TestPatchesAtNode():

    patch_values = np.array([[0, -1, -1, -1],
                             [1,  0, -1, -1],
                             [2,  1, -1, -1],
                             [3,  2, -1, -1],
                             [-1,  3, -1, -1],
                             [4, -1, -1,  0],
                             [5,  4,  0,  1],
                             [6,  5,  1,  2],
                             [7,  6,  2,  3],
                             [-1,  7,  3, -1],
                             [8, -1, -1,  4],
                             [9,  8,  4,  5],
                             [10,  9,  5,  6],
                             [11, 10,  6,  7],
                             [-1, 11,  7, -1],
                             [-1, -1, -1,  8],
                             [-1, -1,  8,  9],
                             [-1, -1,  9, 10],
                             [-1, -1, 10, 11],
                             [-1, -1, 11, -1]])

    patch_mask = np.array([[False,  True,  True,  True],
                           [False, False,  True,  True],
                           [False, False,  True,  True],
                           [False, False,  True,  True],
                           [True, False,  True,  True],
                           [False,  True,  True, False],
                           [False, False, False, False],
                           [False, False, False, False],
                           [False, False, False, False],
                           [True, False, False,  True],
                           [False,  True,  True, False],
                           [False, False, False, False],
                           [False, False, False, False],
                           [False, False, False, False],
                           [True, False, False,  True],
                           [True,  True,  True, False],
                           [True,  True, False, False],
                           [True,  True, False, False],
                           [True,  True, False, False],
                           [True,  True, False,  True]])

    def test_patches_at_node(self):
        rmg = RasterModelGrid((4, 5))
        patches_out = rmg.patches_at_node
        assert_array_equal(patches_out, self.patch_values)

    def test_nodes_at_patch(self):
        rmg = RasterModelGrid((4, 5))
        nodes_out = rmg.nodes_at_patch
        assert_array_equal(nodes_out,
                           np.array([[6,  5,  0,  1],
                                     [7,  6,  1,  2],
                                     [8,  7,  2,  3],
                                     [9,  8,  3,  4],
                                     [11, 10,  5,  6],
                                     [12, 11,  6,  7],
                                     [13, 12,  7,  8],
                                     [14, 13,  8,  9],
                                     [16, 15, 10, 11],
                                     [17, 16, 11, 12],
                                     [18, 17, 12, 13],
                                     [19, 18, 13, 14]]))


class TestSlopesAtPatches():

    def test_slopes_at_patches(self):
        rmg = RasterModelGrid((4, 5))
        rmg.at_node['topographic__elevation'] = rmg.node_x.copy()
        slopes_out = rmg.calc_slope_at_node()
        assert_array_almost_equal(slopes_out, np.full(20, np.pi/4.,
                                  dtype=float))

    def test_slopes_at_patches_comps(self):
        rmg = RasterModelGrid((4, 5))
        rmg.at_node['topographic__elevation'] = -rmg.node_y
        slopes_out = rmg.calc_slope_at_node(
            rmg.at_node['topographic__elevation'],
            return_components=True)
        assert_array_almost_equal(slopes_out[0],
                                  np.full(20, np.pi/4., dtype=float))
        assert_array_almost_equal(slopes_out[1][1],
                                  np.full(20, -np.pi/4., dtype=float))
        assert_array_almost_equal(slopes_out[1][0],
                                  np.zeros(20, dtype=float))
